unit Components;

interface

uses  Math387, CuBlasMtxVec, Classes;

const
   // Now declare some handy global constants
   one : double = 1.0;
   zero : double = 0.0;
   // Boltzman constant
   kBoltzman : double = 1.3806266e-23;
   // Electron charge
   eCharge : double = 1.6021918e-19;
   // Permittivity of free space
   epsilon0 : double = 8.854194e-12;
   // Permeability of free space
   //mu0 : double = pi * 4e-7;
   // Speed of light in free space
   c0 : double = 2.998633e+08;
   // Node Datum
   Datum : Integer = -1;

type
   TElement=Class
   public
   procedure Eval(t: TSample; x, dxdt: TVec); virtual; abstract;
   destructor Destroy; override;
   private
   //  number of terminals
   numterms: integer;
   // Number of parameters of this element
   n_par : integer;
   procedure setNumTerms(numterm: integer);
   end;

   TResistor=Class(TElement)
   public
   constructor Create(pos_terminal,neg_terminal: integer; Value: TSample);
   procedure Eval(t: TSample; x, dxdt: TVec); override;
   function getI: TSample;
   function getp: Integer;
   function getn: Integer;
   destructor Destroy; override;
   private
   p, n : integer;
   // parameters
   res, i, v: TSample;
   end;

   TChuaResistor=Class(TElement)
   public
   constructor Create(pos_terminal,neg_terminal: integer; ValueGa, ValueGB,
   ValueVe : TSample);
   procedure Eval(t: TSample; x, dxdt: TVec); override;
   function getI: TSample;
   function getp: Integer;
   function getn: Integer;
   destructor Destroy; override;
   private
   p, n : integer;
   // parameters
   Ga, Gb, Ve, i: TSample;
   v: TSample;
   end;

   TCapacitor=Class(TElement)
   public
   constructor Create(pos_terminal,neg_terminal: integer; Value: TSample);
   procedure Eval(t: TSample; x, dxdt: TVec); override;
   function getI: TSample;
   function getp: Integer;
   function getn: Integer;
   destructor Destroy; override;
   private
   p, n : integer;
   // parameters
   Cap, i: TSample;
   dvdt: TSample;
   end;

   TInductor=Class(TElement)
   public
   constructor Create(p_terminal,n_terminal: integer; value: TSample);
   procedure Eval(t: TSample; x, dxdt: TVec); override;
   function getV: TSample;
   function getp: Integer;
   function getn: Integer;
   destructor Destroy; override;
   private
   p, n: Integer;
   // parameters
   Ind, v: TSample;
   end;

   TVSinSource=Class(TElement)
   public
   constructor Create(p_terminal,n_terminal: integer; Amp, Freq: TSample);
   procedure Eval(t: TSample; x, dxdt: TVec); override;
   function getV: TSample;
   function getp: Integer;
   function getn: Integer;
   destructor Destroy; override;
   private
   p, n: Integer;
   A,frec,v: TSample;
   end;

   TVABSSinSource=Class(TElement)
   public
   constructor Create(p_terminal,n_terminal: integer; Amp, Freq: TSample);
   procedure Eval(t: TSample; x, dxdt: TVec); override;
   function getV: TSample;
   function getp: Integer;
   function getn: Integer;
   destructor Destroy; override;
   private
   p, n: Integer;
   A,frec,v: TSample;
   end;

   TDCVSource=Class(TElement)
   public
   constructor Create(p_terminal,n_terminal: integer; Voltage: TSample);
   procedure Eval(t: TSample; x, dxdt: TVec); override;
   function getV: TSample;
   function getp: Integer;
   function getn: Integer;
   destructor Destroy; override;
   private
   p, n: Integer;
   v: TSample;
   end;

   TDCISource=Class(TElement)
   public
   constructor Create(p_terminal,n_terminal: integer; Current: TSample);
   procedure Eval(t: TSample; x, dxdt: TVec); override;
   function getI: TSample;
   function getp: Integer;
   function getn: Integer;
   destructor Destroy; override;
   private
   p, n: Integer;
   I: TSample;
   end;
   {Simple diode
    Information
    The simple diode is a one port. It consists of the diode itself and an
    parallel ohmic resistance R. The diode formula is:
                  v/vt
     i  =  ids ( e    - 1).
    If the exponent v/vt reaches the limit maxex, the diode characterisic
    is linearly continued to avoid overflow.}
   TDiodeModelica=Class(TElement)
   public
   constructor Create(p_terminal,n_terminal: integer);
   procedure Eval(t: TSample; x, dxdt: TVec); override;
   function getI: TSample;
   function getp: Integer;
   function getn: Integer;
   destructor Destroy; override;
   private
   p, n: Integer;
   // parameters
   Ids, Vt, Maxexp, R, i: TSample;
   end;

   TDiode=Class(TElement)
   public
   constructor Create(p_terminal,n_terminal: Integer);
   procedure Eval(t: TSample; x, dxdt: TVec); override;
   function getI: TSample;
   function getp: Integer;
   function getn: Integer;
   destructor Destroy; override;
   private
   // Some constants
   p, n: Integer;
   v1, k1, k2, k3, k4, k5, k6: TSample;
   // Parameter variables
   js, alfa, jb, vb, e, ct0, fi, gama, cd0, afac,
   r0, t, area, imax, eg, m, aro, bro, afag, xti, i: TSample;
   end;

   TDiodeZener=Class(TElement)
   public
   constructor Create(p_terminal,n_terminal: Integer; Vz: TSample);
   procedure Eval(t: TSample; x, dxdt: TVec); override;
   function getI: TSample;
   function getp: Integer;
   function getn: Integer;
   destructor Destroy; override;
   private
   p, n: Integer;
   Vzener, Iz, Gd: TSample;
   end;

   {Simple BJT according to Ebers-Moll
   Information
   This model is a simple model of a bipolar npn
   junction transistor according to Ebers-Moll.
   References:
   Vlach, J.; Singal, K.: Computer methods for circuit analysis and design.
   Van Nostrand Reinhold, New York 1983 on page 317 ff.}
   TBJTNPN=Class(TElement)
   public
   constructor Create(b_terminal,c_terminal,e_terminal:integer);
   procedure Eval(t: TSample; x, dxdt: TVec); override;
   function getCi: TSample;
   function getBi: TSample;
   function getEi: TSample;
   function gete: Integer;
   function getc: Integer;
   function getb: integer;
   destructor Destroy; override;
   protected
   vbc, vbe, qbk, ibc, ibe, cbc, cbe,
   ExMin, ExMax, Capcje, Capcjc: TSample;
   private
   Bv, Cv, Ev, dBv, dCv, dEv,
   derVbc, derVbe, derCv: TSample;
   e,b,c: Integer;
   {
   Bf:=50; // "Forward beta";
   Br:=0.1; // "Reverse beta";
   Isat:=1.e-16; // "Transport saturation current";
   Vak:=0.02; // "Early voltage (inverse), 1/Volt";
   Tauf:=0.12e-9; // "Ideal forward transit time";
   Taur:=5e-9; // "Ideal reverse transit time";
   Ccs:=1e-12; // "Collector-substrat(ground) cap.";
   Cje:=0.4e-12; // "Base-emitter zero bias depletion cap.";
   Cjc:=0.5e-12; // "Base-coll. zero bias depletion cap.";
   Phie:=0.8; // "Base-emitter diffusion voltage";
   Me:=0.4; // "Base-emitter gradation exponent";
   Phic:=0.8; // "Base-collector diffusion voltage";
   Mc:=0.333; // "Base-collector gradation exponent";
   Gbc:=1e-15; // "Base-collector conductance";
   Gbe:=1e-15; // "Base-emitter conductance";
   Vt:=0.02585; // "Voltage equivalent of temperature";
   EMin:=-100; // "if x < EMin, the exp(x) function is linearized";
   EMax:=40; // "if x > EMax, the exp(x) function is linearized";}
   Bf, Br, Isat, Vak, Tauf, Taur, Ccs, Cje, Cjc,
   Phie, Me, PHic, Mc, Gbc, Gbe, Vt, Emin, EMax,
   Bi, Ci, Ei: TSample;
   end;

   {Simple BJT according to Ebers-Moll
    Information
    This model is a simple model of a bipolar pnp junction transistor
    according to Ebers-Moll.
    References:
     Vlach, J.; Singal, K.: Computer methods for circuit analysis and design.
     Van Nostrand Reinhold, New York 1983 on page 317 ff.}
   TBJTPNP=Class(TElement)
   public
   constructor Create(b_terminal,c_terminal,e_terminal:integer);
   procedure Eval(t: TSample; x, dxdt: TVec); override;
   function getCi: TSample;
   function getBi: TSample;
   function getEi: TSample;
   function gete: Integer;
   function getc: Integer;
   function getb: integer;
   destructor Destroy; override;
   protected
   vbc, vbe, qbk, ibc, ibe, cbc, cbe,
   ExMin, ExMax, Capcje, Capcjc: TSample;
   private
   e,b,c: Integer;
   {
   Bf:=50; // "Forward beta";
   Br:=0.1; // "Reverse beta";
   Isat:=1.e-16; // "Transport saturation current";
   Vak:=0.02; // "Early voltage (inverse), 1/Volt";
   Tauf:=0.12e-9; // "Ideal forward transit time";
   Taur:=5e-9; // "Ideal reverse transit time";
   Ccs:=1e-12; // "Collector-substrat(ground) cap.";
   Cje:=0.4e-12; // "Base-emitter zero bias depletion cap.";
   Cjc:=0.5e-12; // "Base-coll. zero bias depletion cap.";
   Phie:=0.8; // "Base-emitter diffusion voltage";
   Me:=0.4; // "Base-emitter gradation exponent";
   Phic:=0.8; // "Base-collector diffusion voltage";
   Mc:=0.333; // "Base-collector gradation exponent";
   Gbc:=1e-15; // "Base-collector conductance";
   Gbe:=1e-15; // "Base-emitter conductance";
   Vt:=0.02585; // "Voltage equivalent of temperature";
   EMin:=-100; // "if x < EMin, the exp(x) function is linearized";
   EMax:=40; // "if x > EMax, the exp(x) function is linearized";}
   Bf, Br, Isat, Vak, Tauf, Taur, Ccs, Cje, Cjc,
   Phie, Me, PHic, Mc, Gbc, Gbe, Vt, Emin, EMax,
   Bi, Ci, Ei: TSample;
   Bv, Cv, Ev, dBv, dCv, dEv,
    derVbc, derVbe, derCv: TSample;
   end;

   TBJTEMNPN=Class(TElement)
   public
   constructor Create(b_terminal,c_terminal,e_terminal:integer);
   procedure Eval(t: TSample; x, dxdt: TVec); override;
   function getCi: TSample;
   function getBi: TSample;
   function getEi: TSample;
   function gete: Integer;
   function getc: Integer;
   function getb: integer;
   destructor Destroy; override;
   private
   e,b,c: Integer;
   Io, AF, AR, Gmin, Bv, Cv, Ev: TSample;
   Bi, Ci, Ei: TSample;
   function Iexp(v: TSample): TSample;
   function Ibc(vb,vc,ve: TSample): TSample;
   function Ibe(vb,vc,ve: TSample): TSample;
   end;

   TBJTEMPNP=Class(TElement)
   public
   constructor Create(b_terminal,c_terminal,e_terminal:integer);
   procedure Eval(t: TSample; x, dxdt: TVec); override;
   function getCi: TSample;
   function getBi: TSample;
   function getEi: TSample;
   function gete: Integer;
   function getc: Integer;
   function getb: integer;
   destructor Destroy; override;
   private
   e,b,c: Integer;
   Io, AF, AR, Gmin, Bv, Cv, Ev: TSample;
   Bi, Ci, Ei: TSample;
   function Iexp(v: TSample): TSample;
   function Ibc(vb,vc,ve: TSample): TSample;
   function Ibe(vb,vc,ve: TSample): TSample;
   end;

   TBJTPWLNPN=Class(TElement)
   public
   constructor Create(b_terminal,c_terminal,e_terminal:integer);
   procedure Eval(t: TSample; x, dxdt: TVec); override;
   function getCi: TSample;
   function getBi: TSample;
   function getEi: TSample;
   function gete: Integer;
   function getc: Integer;
   function getb: integer;
   destructor Destroy; override;
   private
   e,b,c: Integer;
   Gi, Go, Bv, Cv, Ev, u, Vbe, Vcesat, beta: TSample;
   Bi, Ci, Ei: TSample;
   function Ibe(vb,vc,ve: TSample): TSample;
   function Ice(vb,vc,ve: TSample): TSample;
   end;

   TBJTPWLPNP=Class(TElement)
   public
   constructor Create(b_terminal,c_terminal,e_terminal:integer);
   procedure Eval(t: TSample; x, dxdt: TVec); override;
   function getCi: TSample;
   function getBi: TSample;
   function getEi: TSample;
   function gete: Integer;
   function getc: Integer;
   function getb: integer;
   destructor Destroy; override;
   private
   e,b,c: Integer;
   Gi, Go, Bv, Cv, Ev, u, Vbe, Vcesat, beta: TSample;
   Bi, Ci, Ei: TSample;
   function Ibe(vb,vc,ve: TSample): TSample;
   function Ice(vb,vc,ve: TSample): TSample;
   end;


   {Simple MOS Transistor
    Information
    The PMOS model is a simple model of a p-channel metal-oxide
    semiconductor FET. It differs slightly from the device used in the
    SPICE simulator. For more details please care for H. Spiro.

    The model does not consider capacitances. A high drain-source resistance
    RDS is included to avoid numerical difficulties.

    References:
    Spiro, H.: Simulation integrierter Schaltungen. R. Oldenbourg Verlag
    Muenchen Wien 1990.}
   TPMOS=Class(TElement)
   public
   D, G, S, B: Integer;
   constructor Create(D_terminal, G_terminal, S_terminal, B_terminal: integer);
   procedure Eval(t: TSample; x, dxdt: TVec); override;
   function getDi: TSample;
   function getGi: TSample;
   function getSi: TSample;
   function getBi: TSample;
   destructor Destroy; override;
   protected
   v, uds, ubs, ugst, ud, us, id, gds: TSample;
   private

   Gi, Si, Di, Bi,
   W, //:=20.0e-6 "Width";
   L, //:=6.0e-6 "Length";
   Beta, //=0.0105e-3 "Transconductance parameter";
   Vt, //=-1.0 "Zero bias threshold voltage";
   K2, //=0.41 "Bulk threshold parameter";
   K5, //=0.839 "Reduction of pinch-off region";
   dW, //=-2.5e-6 "Narrowing of channel";
   dL, //=-2.1e-6 "Shortening of channel";
   RDS: //=1.e+7 "Drain-Source-Resistance";
   TSample;
   end;

   {Simple MOS Transistor
    Information
    The NMos model is a simple model of a n-channel metal-oxide
    semiconductor FET. It differs slightly from the device used in the
    SPICE simulator. For more details please care for H. Spiro.}
   TNMOS=Class(TElement)
   public D, G, S, B: Integer;
   constructor Create(D_terminal, G_terminal, S_terminal, B_terminal: integer);
   procedure Eval(t: TSample; x, dxdt: TVec); override;
   function getDi: TSample;
   function getGi: TSample;
   function getSi: TSample;
   function getBi: TSample;
   destructor Destroy; override;
   protected
   v, uds, ubs, ugst, ud, us, id, gds: TSample;
   private
   Gi, Si, Di, Bi,
   W, //:=20.0e-6 "Width";
   L, //:=6.0e-6 "Length";
   Beta, //=0.0105e-3 "Transconductance parameter";
   Vt, //=-1.0 "Zero bias threshold voltage";
   K2, //=0.41 "Bulk threshold parameter";
   K5, //=0.839 "Reduction of pinch-off region";
   dW, //=-2.5e-6 "Narrowing of channel";
   dL, //=-2.1e-6 "Shortening of channel";
   RDS: //=1.e+7 "Drain-Source-Resistance";
   TSample;
   end;

   {Simple nonideal model of an OpAmp with limitation
   Information
   The OpAmp is a simle nonideal model with a smooth out.v = f(vin)
   characteristic, where "vin = in_p.v - in_n.v". The characteristic is
   limited by VMax.v and VMin.v. Its slope at vin=0 is the parameter Slope,
   which must be positive. (Therefore, the absolute value of Slope is
   taken into calculation.)}
   TOPAMPIDEAL=Class(TElement)
   public
   constructor Create(in_pos, in_neg, out_terminal: Integer);
   procedure Eval(t: TSample; x, dxdt: TVec); override;
   function getinp: Integer;
   function getinn: Integer;
   function getout: Integer;
   private
   p, // "Positive pin of the input port";
   n, // "Negative pin of the input port";
   o: Integer; // "Output pin";
   end;

   TOPAMP=Class(TElement)
   public
   constructor Create(in_pos, in_neg, out_terminal: Integer);
   procedure Eval(t: TSample; x, dxdt: TVec); override;
   function getV: TSample;
   function getinp: Integer;
   function getinn: Integer;
   function getout: Integer;
   protected
   f, // "auxiliary variable";
   absSlope: TSample;
   private
   p, // "Positive pin of the input port";
   n, // "Negative pin of the input port";
   o: Integer; // "Output pin";
   Slope: Tsample; // "Slope of the out.v/vin characteristic at vin=0";
   VMaxo,  // "Positive output voltage limitation";
   VMino,  //  "Negative output voltage limitation";
   IMaxo,  IMino,
   vin,   //"input voltagae";
   outv: TSample; //"Output voltaje:;
   end;

   TOPAMP741=Class(TElement)
   public
   constructor Create(inp, inn, output, vcc, vee: integer);
   procedure Eval(t: TSample; x, dxdt: TVec); override;
   function getinp: Integer;
   function getinn: Integer;
   function getout: Integer;
   function getvcc: Integer;
   function getvee: Integer;
   private
   p,n,o,ve,vc: Integer;
   end;

   TOPAMPMM=Class(TElement)
   public
   constructor Create(inp, inn, output, vcc, vee: integer);
   procedure Eval(t: TSample; x, dxdt: TVec); override;
   function getinp: Integer;
   function getinn: Integer;
   function getout: Integer;
   function getvcc: Integer;
   function getvee: Integer;
   private
   p,n,o,ve,vc: Integer;
   end;

   TOPAMPMacroModel=Class(TElement)
   public
   constructor Create(inp, inn, output, vcc, vee: integer);
   procedure Eval(t: TSample; x, dxdt: TVec); override;
   function getinp: Integer;
   function getinn: Integer;
   function getout: Integer;
   function getvcc: Integer;
   function getvee: Integer;
   private
   p,n,o,ve,vc: Integer;
   end;
   {"bjtnpn",
	 "Gummel and Poon model",
   "Senthil Velu",}
   TBJTNPNF=Class(TElement)
   public
   constructor Create(b_terminal,c_terminal,e_terminal:integer);
   procedure Eval(t: TSample; x, dxdt: TVec); override;
   function getCi: TSample;
   function getBi: TSample;
   function getEi: TSample;
   function gete: Integer;
   function getc: Integer;
   function getb: integer;
   destructor Destroy; override;
   private
   // Terminales
   b,c,e : Integer;
   // Current Terminals
   Bi, Ci, Ei: TSample;
   // Parameter variables
   isat, bf, nf, vaf, ikf, ise, ne, br, nr, varr, ikr, isc,
   nc, re, rb, rbm, irb, rc, eg, cje, vje, mje, cjc, vjc, mjc,
   xcjc, fc, tf, xtf, vtf, itf, tr, xtb, xti, tre1, tre2, trb1, trb2,
   trm1, trm2, trc1, trc2, tnom, t, cjs, mjs, vjs, area, ns, iss : TSample;
   {"is","Transport saturation current (A)",
	 { "bf","Ideal maximum forward beta",
   {"nf","Forward current emission coefficient"
	 {"vaf","Forward early voltage (V)",

	 {"ikf","Forward-beta high current roll-off knee current (A)"
	{"ise","Base-emitter leakage saturation current (A)"
	{"ne","Base-emitter leakage emission coefficient"
	{"br","Ideal maximum reverse beta"
	{"nr","Reverse current emission coefficient"
	{"var","revers early voltage (V)"

	{"ikr","Corner for reverse-beta high current roll off (A)"
	{"isc","Base collector leakage saturation current (A)"
	{"nc","Base-collector leakage emission coefficient"
	{"re","Emitter ohmic resistance (W)"
	{"rb","Zero bias base resistance (W)"
	{"rbm","Minimum base resistance (W)"
	{"irb","Current at which rb falls to half of rbm (A)"
	{"rc","Collector ohmic resistance (W)"
	{"eg","Badgap voltage (eV)"
	{"cje","Base emitter zero bias p-n capacitance (F)"
	{"vje","Base emitter built in potential (V)"
	{"mje","Base emitter p-n grading factor"
	{"cjc","Base collector zero bias p-n capacitance (F)"
	{"vjc","Base collector built in potential (V)"
	{"mjc","Base collector p-n grading factor"
	{"xcjc","Fraction of cbc connected internal to rb"
	{"fc","Forward bias depletion capacitor coefficient"
	{"tf","Ideal forward transit time (S)"
	{"xtf","Transit time bias dependence coefficient"
	{"vtf","Transit time dependency on vbc (V)"
	{"itf","Transit time dependency on ic (A)"
	{"tr","Ideal reverse transit time (S)"
	{"xtb","Forward and reverse beta temperature coefficient"
	{"xti","IS temperature effect exponent"
	{"tre1","RE temperature coefficient (linear)"
	{"tre2","RE temperature coefficient (quadratic)"
	{"trb1","RB temperature coefficient (linear)"
	{"trb2","RB temperature coefficient  (quadratic)"
	{"trm1","RBM temperature coefficient (linear)"
	{"trm2","RBM temperature coefficient (quadratic)"
	{"trc1","RC temperature coefficient (linear)"
	{"trc2","RC temperature coefficient (quadratic)"
	{"tnom","Nominal temperature (K)"
	{"t","temperature (K)"
	{"cjs", "Collector substrate capacitance"
	{"mjs", "substrate junction exponential factor"
	{"vjs", "substrate junction built in potential"
	{"area","Current multiplier"
	{"ns","substrate p-n coefficient"
	{"iss","Substrate saturation current"}
  IIbe,IIbc,IIce,Ibf,Ile,Ibr,Ilc,kqb,kqbtemp,
	  cbej,cbet,cbcj,cbct,ibe,ibc, vth,
    Ibf1, Ibr1, kqbtem, tempvaf, tempvar, tZ,
    Tibe, Tibc, rb1, vbx, cbx, ibx,
    dibc, dibe, dIIbc, dIIbe, IIjs, cjjs, ijs,
    Bv, Cv, Ev, dBv, dCv, dEv, vbc, vbe,
    dVbcdt, dVbedt, dVcdt,
    Vcjs, dvcjsdt, d2vbedt, d2vbcdt: TSample;
   end;

   {"bjtpnp",
	 "Gummel and Poon model",
   "Senthil Velu",}
   TBJTPNPF=Class(TElement)
   public
   constructor Create(b_terminal,c_terminal,e_terminal:integer);
   procedure Eval(t: TSample; x, dxdt: TVec); override;
   function getCi: TSample;
   function getBi: TSample;
   function getEi: TSample;
   function gete: Integer;
   function getc: Integer;
   function getb: integer;
   destructor Destroy; override;
   private
   // Terminales
   b,c,e : Integer;
   // Current Terminals
   Bi, Ci, Ei: TSample;
   // Parameter variables
   isat, bf, nf, vaf, ikf, ise, ne, br, nr, varr, ikr, isc,
   nc, re, rb, rbm, irb, rc, eg, cje, vje, mje, cjc, vjc, mjc,
   xcjc, fc, tf, xtf, vtf, itf, tr, xtb, xti, tre1, tre2, trb1, trb2,
   trm1, trm2, trc1, trc2, tnom, t, cjs, mjs, vjs, area, ns, iss : TSample;
   {"is","Transport saturation current (A)",
	 { "bf","Ideal maximum forward beta",
   {"nf","Forward current emission coefficient"
	 {"vaf","Forward early voltage (V)",

	 {"ikf","Forward-beta high current roll-off knee current (A)"
	{"ise","Base-emitter leakage saturation current (A)"
	{"ne","Base-emitter leakage emission coefficient"
	{"br","Ideal maximum reverse beta"
	{"nr","Reverse current emission coefficient"
	{"var","revers early voltage (V)"

	{"ikr","Corner for reverse-beta high current roll off (A)"
	{"isc","Base collector leakage saturation current (A)"
	{"nc","Base-collector leakage emission coefficient"
	{"re","Emitter ohmic resistance (W)"
	{"rb","Zero bias base resistance (W)"
	{"rbm","Minimum base resistance (W)"
	{"irb","Current at which rb falls to half of rbm (A)"
	{"rc","Collector ohmic resistance (W)"
	{"eg","Badgap voltage (eV)"
	{"cje","Base emitter zero bias p-n capacitance (F)"
	{"vje","Base emitter built in potential (V)"
	{"mje","Base emitter p-n grading factor"
	{"cjc","Base collector zero bias p-n capacitance (F)"
	{"vjc","Base collector built in potential (V)"
	{"mjc","Base collector p-n grading factor"
	{"xcjc","Fraction of cbc connected internal to rb"
	{"fc","Forward bias depletion capacitor coefficient"
	{"tf","Ideal forward transit time (S)"
	{"xtf","Transit time bias dependence coefficient"
	{"vtf","Transit time dependency on vbc (V)"
	{"itf","Transit time dependency on ic (A)"
	{"tr","Ideal reverse transit time (S)"
	{"xtb","Forward and reverse beta temperature coefficient"
	{"xti","IS temperature effect exponent"
	{"tre1","RE temperature coefficient (linear)"
	{"tre2","RE temperature coefficient (quadratic)"
	{"trb1","RB temperature coefficient (linear)"
	{"trb2","RB temperature coefficient  (quadratic)"
	{"trm1","RBM temperature coefficient (linear)"
	{"trm2","RBM temperature coefficient (quadratic)"
	{"trc1","RC temperature coefficient (linear)"
	{"trc2","RC temperature coefficient (quadratic)"
	{"tnom","Nominal temperature (K)"
	{"t","temperature (K)"
	{"cjs", "Collector substrate capacitance"
	{"mjs", "substrate junction exponential factor"
	{"vjs", "substrate junction built in potential"
	{"area","Current multiplier"
	{"ns","substrate p-n coefficient"
	{"iss","Substrate saturation current"}
  IIbe,IIbc,IIce,Ibf,Ile,Ibr,Ilc,kqb,kqbtemp,
	  cbej,cbet,cbcj,cbct,ibe,ibc, vth,
    Ibf1, Ibr1, kqbtem, tempvaf, tempvar, tZ,
    Tibe, Tibc, rb1, vbx, cbx, ibx,
    dibc, dibe, dIIbc, dIIbe, IIjs, cjjs, ijs,
    Bv, Cv, Ev, dBv, dCv, dEv, vbc, vbe,
    dVbcdt, dVbedt, dVcdt,
    Vcjs, dvcjsdt, d2vbedt, d2vbcdt: TSample;
   end;

   TVCT=Class(TElement)
   public
   constructor Create(inpos_terminal, inneg_terminal, outpos_terminal, outneg_terminal: Integer;
   G: TSample);
   procedure Eval(t: TSample; x, dxdt: TVec); override;
   function getI: TSample;
   function getinp: Integer;
   function getinn: Integer;
   function getoutp: Integer;
   function getoutn: Integer;
   private
   pin, nin, pout, nout: Integer;
   Gtran, I: TSample;
   end;

   TVVT=Class(TElement)
   public
   constructor Create(inpos_terminal, inneg_terminal, outpos_terminal, outneg_terminal: Integer;
   Mu: TSample);
   procedure Eval(t: TSample; x, dxdt: TVec); override;
   function getV: TSample;
   function getinp: Integer;
   function getinn: Integer;
   function getoutp: Integer;
   function getoutn: Integer;
   private
   pin, nin, pout, nout: Integer;
   Mutran, V: TSample;
   end;


   procedure condassign(var a: double; b,c,d: Double);

implementation

uses Math;

procedure condassign(var a: double; b,c,d: Double);
 begin
    if b>0 then
    a:=c
    else
    a:=d;
 end;

   { TElement }

destructor TElement.Destroy;
begin
   inherited;
end;

procedure TElement.setNumTerms(numterm: integer);
begin
  numterms:= numterm;
end;

{ TDiode }

constructor TDiode.Create(p_terminal,n_terminal: Integer);
begin
  p:=p_terminal-1;
  n:=n_terminal-1;
  v1 := log10(5e8 / alfa) / alfa; // normal is .5e9
  k1 := ct0 / power(0.2, gama);
  k2 := fi * 0.8;
  k3 := exp(alfa * v1);
  k4 := - ct0 * fi / (one - gama);
  k5 := k4 * (0.2 * k1 / ct0 - one);
  k6 := cd0 / afac;
  n_par := 20;
  js := 1e-16;
  alfa := 38.696;
  jb := 1e-5;
  vb := -1e20;
  e := 10.0;
  ct0 := zero;
  fi := 0.8;
  gama := 0.5;
  cd0 := zero;
  afac := 38.696;
  r0 := zero;
  t := zero;
  area := one;
  imax := zero;
  eg := 0.8;
  m := 0.5;
  aro := zero;
  bro := zero;
  afag := one;
  xti := 2.0;
  // Set the number of terminals
  setNumTerms(2);
end;

procedure TDiode.Eval(t: TSample; x, dxdt: TVec);
var vj, dvj_dx, cj, rs, itmp,
    exp1, k14, k15, v , dvdt: TSample;
begin
  // x[0]: state variable
  // x[1]: time derivative of x[0]
  v:=0.0; dvdt:=0.0;
  if (p<>Datum) then  begin
      v:=x[p]; dvdt:=dxdt[p]; end;
  if n<>Datum then begin
      v:=v-x[n]; dvdt:=dvdt-dxdt[n]; end;


  condassign(vj, v1 - v,
	     x[0] + zero,
	     v1 + log10(one + alfa*(v-v1))/alfa);
  condassign(dvj_dx, v1 - v,
	     one,
	     one / (one + alfa*(v-v1)));
  // Calculate the junction capacitance (experimental)
   // Use a modified function with continous first and second deriv.
   cj := zero;
   if ct0<>0 then
   begin
     exp1 := exp(10.0 * (vj-k2));
     k14 := ct0 * gama / fi;
     k15 := k4 * (gama - one) / fi;
     condassign(cj, -vj,
 	       ct0 / power(one - vj/fi, gama),
 	       (ct0 + vj * (k14 + k15 * vj)) / (one + exp1) +
 	       k1 * exp1 / (one + exp1) );
   end;
   //    // Calculate capacitance as usual
   //    cj = zero;
   //    if (isSet(&ct0))
   //      condassign(cj, vj - k2,
   //  	       k1,
   //  	       ct0 / pow(one - vj/fi, gama));

  if cd0<>0 then
    // add depletion capacitance
    cj := cj + cd0 * exp(afac * vj);

  // Now calculate the current through the capacitor.
  // Using the chain rule:
  //
  // dq/dt = cj(vj) * dvj/dx * dx/dt
  // x[1] is dx/dt
  i := cj * dvj_dx * dvdt;

  // Now use the state variable again to calculate the total
  // current.  This way, we save some exp() calls.  The total
  // current is the current through the capacitor plus the ideal
  // diode current.
  condassign(itmp, v1 - v,
	     js * (exp(alfa * v) - one),
	     js * k3 * (one + alfa * (v - v1)) - js);
  i := i + itmp;

  // substract the breakdown current
  condassign(itmp, vj - vb - one,
	     zero ,
	     jb * power(one + vb - vj, e));
  i := i - itmp;

  // Calculate Rs
  if (cj <> zero) then
    condassign(rs, -r0 + t/cj,
	       zero , r0 - t/cj)
  else
    rs := r0;

  {vp[0] := vj + ip[0] * rs;}

  // scale the current according to area. All the calculations were made
  // for a unit area diode.
  i := i * area;
end;

destructor TDiode.Destroy;
begin
  inherited;
end;

function TDiode.getn: Integer;
begin
  result:= n;
end;

function TDiode.getp: Integer;
begin
  result:= p;
end;

{ TBJT }

constructor TBJTNPNF.Create(b_terminal,c_terminal,e_terminal:integer);
begin
  b:=b_terminal-1;
  c:=c_terminal-1;
  e:=e_terminal-1;
  isat := 1e-16;
	bf := 100.0;
  nf := one;
	vaf := zero;
	ikf := zero;
	ise := zero;
	ne := 1.5;
  br := one;
	nr := one;
	varr := zero;
	ikr := zero;
	isc := zero;
	nc := 2.0;
	re := zero;
	rb := zero;
	rbm := zero;
	irb := zero;
	rc := zero;
	eg := 1.11;
	cje := zero;
	vje := 0.75;
	mje := 0.33;
	cjc := zero;
	vjc := 0.75;
	mjc := 0.33;
	xcjc := one;
	fc := 0.5;
	tf := zero;
	xtf := zero;
	vtf := zero;
	itf := zero;
	tr := zero;
	xtb := zero;
	xti := 3.0;
	tre1 := zero;
	tre2 := zero;
	trb1 := zero;
	trb2 := zero;
	trm1 := zero;
	trm2 := zero;
	trc1 := zero;
	trc2 := zero;
	tnom := 300.0;
	t := 300.0;
	cjs := zero;
	mjs := zero;
	vjs := 0.75;
	area := one;
	ns := one;
	iss := zero;

	setNumTerms(4);
end;

destructor TBJTNPNF.Destroy;
begin

end;

procedure TBJTNPNF.Eval(t: TSample; x, dxdt: TVec);
begin
  if b=datum then begin Bv:=0.0; dBv:=0.0; end
  else begin Bv:=x[b]; dBv:=dxdt[b]; end;
  if c=datum then begin Cv:=0.0; dCv:=0.0; end
  else begin Cv:=x[c]; dCv:=dxdt[c]; end;
  if e=datum then begin Ev:=0.0; dEv:=0.0; end
  else begin Ev:=x[e]; dEv:=dxdt[e]; end;
  vbc := Bv - Cv;
  vbe := Bv - Ev;
  dvbcdt:= dBv-dCv;
  dvbedt:= dBv-dEv;
  dvCdt:= dCv;
  Vcjs:=Cv;
  dvcjsdt:=dvCdt;
  d2vbedt:=0.0;
  d2vbcdt:=0.0;
  //x[vbe,vbc,vcjs,dvbe,dvbc,dvcjs,d2vbe,d2vbc,d2vcjs]
	//x[0] : Vbe   Vbe
	//x[1] : Vbc   Vbc
	//x[2] : Vcjs
	//x[3] : dvbe/dt  dvbedt
	//x[4] : dvbc/dt  dvbcdt
	//x[5] : dvcjs/dt
	//x[6] : d2vbe/dt
	//x[7] : d2vbc/dt
	//x[8] : d2vcjs/dt
	//vp[0] : Vcjs ip[0] : Ic
	//vp[1] : Vbjs ip[1] : Ib
	//vp[2] : Vejs ip[2] : Ie
  vth := (kBoltzman * tnom) / eCharge;
  if (rbm<>0) then
    rbm := rb;

  //dc current equations

	Ibf := isat * (exp(Vbe/nf/vth) - one);
	Ile := ise * (exp(Vbe/ne/vth) - one);
	Ibr := isat * (exp(Vbc/nr/vth) - one);
	Ilc := isc * (exp(Vbc/nc/vth) - one);
	IIbe := Ibf / bf + Ile;
	IIbc := Ibr / br + Ilc;

  Ibf1 := (IIbe - Ile) * bf;
	Ibr1 := (IIbc - Ilc) * br;
	kqbtem := zero;

  if (ikf<>0) then
		kqbtem :=  4.0*(Ibf1/ikf);
	if (ikr<>0) then
		kqbtem := kqbtem + 4.0*(Ibr1/ikr);
	kqbtemp := sqrt(one + kqbtem);

  tempvaf := zero;
  if (vaf<>0) then
		tempvaf := Vbc / vaf;
  tempvar := zero;
	if (varr<>0) then
		tempvar := Vbe / varr;

  kqb := 0.5 * (one / (one - tempvaf - tempvar)) * (one + kqbtemp);
	IIce := (Ibf1 - Ibr1) / kqb;

  // Charge base-collector
	condassign(cbcj,Vbc-fc*vjc,
	cjc*power(one-fc,-one-mjc)*(one-fc*(one+mjc)+mjc*Vbe/vjc),
	cjc*power(one-Vbc/vjc,-mjc));
	cbct:=tr*isat*exp(Vbc/nr/vth)/(nr*vth);
	ibc:=area*(cbct+xcjc*cbcj)*dvbcdt;
	//end

  //Current Base-Emitter
	condassign(cbej,Vbe-fc*vje,
	cjc*power(one-fc,-one-mjc)*(one-fc*(one+mje)+mje*Vbe/vje),
	cje*power(one-Vbe/vje,-mje));

  tZ := one;
	if (vtf<>0) then
		tZ := exp(0.69*Vbc/vtf);

	cbet:=isat*exp(Vbe/nf/vth)*tf*(one+xtf*tZ)/(nf*vth);
  ibe:=area*(cbet+cbej)*dvbedt;

  //end

  Tibe := IIbe + ibe;
	Tibc := IIbc + ibc;

  rb1 := (rbm + (rb - rbm) / kqb) / area;
	vbx := Vbc + (Tibe + Tibc) * rb;

  condassign(cbx,vbx-fc*vjc,
	(one-xcjc)*cjc*power(one-fc,-one-mjc)*(one-fc*(one+mjc)+mjc*vbx/vjc),
	(one-xcjc)*cjc*power(one-vbx/vjc,-mjc));

  dibc := area*(cbct+xcjc*cbcj) * d2vbcdt;
	dibe := area*(cbet+cbej) * d2vbedt;
	dIIbe := isat/bf * exp(Vbe/nf/vth) * (dvbedt/nf/vth)
	+ ise * exp(Vbe/ne/vth) * (dvbedt/ne/vth);
	dIIbc := isat/br * exp(Vbe/nr/vth) * (dvbcdt/nr/vth)
	+ isc * exp(Vbc/nc/vth) * (dvbcdt/nc/vth);

	ibx := cbx * (dvbedt + rb*(dibc+dibe+dIIbc+dIIbe));

  IIjs := area * iss * (exp(Vcjs/ns/vth) - one);

  condassign(cjjs, Vcjs - zero,
	cjs*(one+mjs*Vcjs/vjs),
	cjs*power(one-Vcjs/vjs,-mjs));
	ijs := IIjs + area*(cjjs * dvcjsdt);

  Ci := IIce - Tibc - ibx - ijs; // collector current
	Bi := ibx + Tibc + Tibe;      // base current
	Ei := -(IIce + Tibe - ijs);    // emitter current

	{vp[0] := x[2] + rc*ip[0];                // Collector Substrate Voltage
	vp[1] := x[2] + x[1] + ip[1]*rb;        // Base Substrate Voltage
	vp[2] := x[1] - x[0] + x[2] + ip[2]*re; // Emitter Substrate Voltage}
end;

function TBJTNPNF.getb: integer;
begin
  result:= b;
end;

function TBJTNPNF.getBi: TSample;
begin
   Result:=Bi;
end;

function TBJTNPNF.getc: Integer;
begin
   result:= c;
end;

function TBJTNPNF.getCi: TSample;
begin
   Result:=Ci;
end;

function TBJTNPNF.gete: Integer;
begin
   result:= e;
end;

function TBJTNPNF.getEi: TSample;
begin
   Result:=Ei;
end;

function TDiode.getI: TSample;
begin
  result:=i;
end;

{ TDiodeModelica }

constructor TDiodeModelica.Create(p_terminal, n_terminal: integer);
begin
   p:=p_terminal-1;
   n:=n_terminal-1;
   Ids:=1.0e-6; //"Saturation current";
   Vt:=0.04; //"Voltage equivalent of temperature (kT/qn)";
   Maxexp := 15; //(final min=Modelica.Constants.SMALL)
                 //"Max. exponent for linear continuation";
   R:=1.0e8; //"Parallel ohmic resistance";
end;

destructor TDiodeModelica.destroy;
begin

  inherited;
end;

procedure TDiodeModelica.Eval(t: TSample; x, dxdt: TVec);
var v: TSample;
begin
   v:= 0.0;
   if (p<>Datum) then
      v:=x[p];
   if n<>Datum then
      v:=v-x[n];
   if (v/Vt > Maxexp) then
   i:=Ids*(exp(Maxexp)*(1 + v/Vt - Maxexp) - 1) + v/R
   else
   i:=Ids*(exp(v/Vt) - 1) + v/R;
end;

function TDiodeModelica.getI: TSample;
begin
   result:=i;
end;

function TDiodeModelica.getn: Integer;
begin
  result:= n;
end;

function TDiodeModelica.getp: Integer;
begin
  result:= p;
end;

{ TResistence }

constructor TResistor.Create(pos_terminal, neg_terminal: integer;
  Value: TSample);
begin
   p:=pos_terminal-1;
   n:=neg_terminal-1;
   res:=Value;
end;

destructor TResistor.destroy;
begin

  inherited;
end;

procedure TResistor.Eval(t: TSample; x, dxdt: TVec);
begin
   v:= 0.0;
   if (p <> Datum) then
      v:=x[p];
   if (n<>Datum) then
      v:=v-x[n];
   if res<>0.0 then
     i:=v/res
   else
     i:=v/1e-7;  
end;

function TResistor.getI: TSample;
begin
  result:=i;
end;

function TResistor.getn: Integer;
begin
  result:= n;
end;

function TResistor.getp: Integer;
begin
  result:= p;
end;

{ TCapacitor }

constructor TCapacitor.Create(pos_terminal, neg_terminal: integer;
  Value: TSample);
begin
   p:=pos_terminal-1;
   n:=neg_terminal-1;
   cap:=Value;
end;

destructor TCapacitor.destroy;
begin
  inherited;
end;

procedure TCapacitor.Eval(t: TSample; x, dxdt: TVec);
begin
  dvdt:= 0.0;
  if (p <> Datum) then
      dvdt:=dxdt[p];
  if (n <> Datum) then
      dvdt:=dvdt-dxdt[n];
  i:=Cap*dvdt;
end;

function TCapacitor.getI: TSample;
begin
   result:=i;
end;

function TCapacitor.getn: Integer;
begin
  result:= n;
end;

function TCapacitor.getp: Integer;
begin
  result:= p;
end;

{ TInductor }

constructor TInductor.Create(p_terminal, n_terminal: integer;
  value: TSample);
begin
  p:= P_terminal-1;
  n:= n_terminal-1;
  ind:= value;
end;

destructor TInductor.destroy;
begin
  inherited;
end;

procedure TInductor.Eval(t: TSample; x, dxdt: TVec);
var didt: TSample;
begin
  didt:=0.0;
  if (p<>Datum) then
      didt:=dxdt[p];
   if n<>Datum then
      didt:=didt-dxdt[n];
   v:=Ind*didt;
end;

function TInductor.getV: TSample;
begin
   result:=v;
end;

function TInductor.getn: Integer;
begin
  result:= n;
end;

function TInductor.getp: Integer;
begin
  result:= p;
end;

{ TBJTNPN }

constructor TBJTNPN.Create(b_terminal, c_terminal, e_terminal: integer);
begin
   b:=b_terminal-1;
   c:=c_terminal-1;
   e:=e_terminal-1;
   Bf:=100; // "Forward beta";
   Br:=0.1; // "Reverse beta";
   Isat:=1.e-16; // "Transport saturation current";
   Vak:=0.02; // "Early voltage (inverse), 1/Volt";
   Tauf:=0.12e-9; // "Ideal forward transit time";
   Taur:=5e-9; // "Ideal reverse transit time";
   Ccs:=1e-12; // "Collector-substrat(ground) cap.";
   Cje:=0.4e-12; // "Base-emitter zero bias depletion cap.";
   Cjc:=0.5e-12; // "Base-coll. zero bias depletion cap.";
   Phie:=0.8; // "Base-emitter diffusion voltage";
   Me:=0.4; // "Base-emitter gradation exponent";
   Phic:=0.8; // "Base-collector diffusion voltage";
   Mc:=0.333; // "Base-collector gradation exponent";
   Gbc:=1e-15; // "Base-collector conductance";
   Gbe:=1e-15; // "Base-emitter conductance";
   Vt:=0.02585; // "Voltage equivalent of temperature";
   EMin:=-100; // "if x < EMin, the exp(x) function is linearized";
   EMax:=40; // "if x > EMax, the exp(x) function is linearized";
end;

destructor TBJTNPN.Destroy;
begin

  inherited;
end;

procedure TBJTNPN.Eval(t: TSample; x, dxdt: TVec);
begin
  if b=datum then begin Bv:=0.0; dBv:=0.0; end
  else begin Bv:=x[b]; dBv:=dxdt[b]; end;
  if c=datum then begin Cv:=0.0; dCv:=0.0; end
  else begin Cv:=x[c]; dCv:=dxdt[c]; end;
  if e=datum then begin Ev:=0.0; dEv:=0.0; end
  else begin Ev:=x[e]; dEv:=dxdt[e]; end;
  vbc := Bv - Cv;
  vbe := Bv - Ev;
  dervbc:= dBv-dCv;
  dervbe:= dBv-dEv;
  derCv:= dCv;
  qbk := 1 - vbc*Vak;
  ExMin := exp(EMin);
  ExMax := exp(EMax);
  if (vbc/Vt < EMin) then
    ibc:= Isat*(ExMin*(vbc/Vt - EMin + 1) - 1) + vbc*Gbc
  else if (vbc/Vt > EMax) then
    ibc:= Isat*(ExMax*(vbc/Vt - EMax + 1) - 1) + vbc*Gbc
  else
    ibc:= Isat*(exp(vbc/Vt) - 1) + vbc*Gbc;
  if (vbe/Vt < EMin) then
    ibe := Isat*(ExMin*(vbe/Vt - EMin + 1) - 1) + vbe*Gbe
  else if (vbe/Vt > EMax) then
    ibe := Isat*(ExMax*(vbe/Vt - EMax + 1) - 1) + vbe*Gbe
  else
    ibe := Isat*(exp(vbe/Vt) - 1) + vbe*Gbe;

  if (vbc/Phic > 0) then
    Capcjc := Cjc*(1 + Mc*vbc/Phic)
  else
    Capcjc := Cjc*power(1 - vbc/Phic, -Mc);
  if (vbe/Phie > 0) then
    Capcje := Cje*(1 + Me*vbe/Phie)
  else
    Capcje := Cje*power(1 - vbe/Phie, -Me);
  if (vbc/Vt < EMin) then
    cbc := Taur*Isat/Vt*ExMin*(vbc/Vt - EMin + 1) + Capcjc
  else if (vbc/Vt > EMax) then
    cbc := Taur*Isat/Vt*ExMax*(vbc/Vt - EMax + 1) + Capcjc
  else
    cbc := Taur*Isat/Vt*exp(vbc/Vt) + Capcjc;
  if (vbe/Vt < EMin) then
    cbe := Tauf*Isat/Vt*ExMin*(vbe/Vt - EMin + 1) + Capcje
  else if (vbe/Vt > EMax) then
    cbe := Tauf*Isat/Vt*ExMax*(vbe/Vt - EMax + 1) + Capcje
  else
    cbe := Tauf*Isat/Vt*exp(vbe/Vt) + Capcje;

  Ci := (ibe - ibc)*qbk - ibc/Br - cbc*dervbc + Ccs*derCv;
  Bi := ibe/Bf + ibc/Br + cbc*dervbc + cbe*dervbe;
  Ei := -Bi - Ci + Ccs*derCv;
end;

function TBJTNPN.getb: integer;
begin
   result:= b;
end;

function TBJTNPN.getBi: TSample;
begin
   result:=Bi;
end;

function TBJTNPN.getc: Integer;
begin
  result:= c;
end;

function TBJTNPN.getCi: TSample;
begin
   result:=Ci;
end;

function TBJTNPN.gete: Integer;
begin
   result:= e;
end;

function TBJTNPN.getEi: TSample;
begin
   result:=Ei;
end;

{ TBJTPNP }

constructor TBJTPNP.Create(b_terminal, c_terminal, e_terminal: integer);
begin
   b:=b_terminal-1;
   c:=c_terminal-1;
   e:=e_terminal-1;
   Bf:=50; // "Forward beta";
   Br:=0.1; // "Reverse beta";
   Isat:=1.e-16; // "Transport saturation current";
   Vak:=0.02; // "Early voltage (inverse), 1/Volt";
   Tauf:=0.12e-9; // "Ideal forward transit time";
   Taur:=5e-9; // "Ideal reverse transit time";
   Ccs:=1e-12; // "Collector-substrat(ground) cap.";
   Cje:=0.4e-12; // "Base-emitter zero bias depletion cap.";
   Cjc:=0.5e-12; // "Base-coll. zero bias depletion cap.";
   Phie:=0.8; // "Base-emitter diffusion voltage";
   Me:=0.4; // "Base-emitter gradation exponent";
   Phic:=0.8; // "Base-collector diffusion voltage";
   Mc:=0.333; // "Base-collector gradation exponent";
   Gbc:=1e-15; // "Base-collector conductance";
   Gbe:=1e-15; // "Base-emitter conductance";
   Vt:=0.02585; // "Voltage equivalent of temperature";
   EMin:=-100; // "if x < EMin, the exp(x) function is linearized";
   EMax:=40; // "if x > EMax, the exp(x) function is linearized";
end;

destructor TBJTPNP.Destroy;
begin

  inherited;
end;

procedure TBJTPNP.Eval(t: TSample; x, dxdt: TVec);
begin
  if b=datum then begin Bv:=0.0; dBv:=0.0; end
  else begin Bv:=x[b]; dBv:=dxdt[b]; end;
  if c=datum then begin Cv:=0.0; dCv:=0.0; end
  else begin Cv:=x[c]; dCv:=dxdt[c]; end;
  if e=datum then begin Ev:=0.0; dEv:=0.0; end
  else begin Ev:=x[e]; dEv:=dxdt[e]; end;
  vbc := Cv - Bv;
  vbe := Ev - Bv;
  dervbc:= dCv-dBv;
  dervbe:= dEv-dBv;
  derCv:= dCv;
  qbk := 1 - vbc*Vak;
  ExMin := exp(EMin);
  ExMax := exp(EMax);
  if (vbc/Vt < EMin) then
    ibc:= Isat*(ExMin*(vbc/Vt - EMin + 1) - 1) + vbc*Gbc
  else if (vbc/Vt > EMax) then
    ibc:= Isat*(ExMax*(vbc/Vt - EMax + 1) - 1) + vbc*Gbc
  else
    ibc:= Isat*(exp(vbc/Vt) - 1) + vbc*Gbc;
  if (vbe/Vt < EMin) then
    ibe := Isat*(ExMin*(vbe/Vt - EMin + 1) - 1) + vbe*Gbe
  else if (vbe/Vt > EMax) then
    ibe := Isat*(ExMax*(vbe/Vt - EMax + 1) - 1) + vbe*Gbe
  else
    ibe := Isat*(exp(vbe/Vt) - 1) + vbe*Gbe;

  if (vbc/Phic > 0) then
    Capcjc := Cjc*(1 + Mc*vbc/Phic)
  else
    Capcjc := Cjc*power(1 - vbc/Phic, -Mc);
  if (vbe/Phie > 0) then
    Capcje := Cje*(1 + Me*vbe/Phie)
  else
    Capcje := Cje*power(1 - vbe/Phie, -Me);
  if (vbc/Vt < EMin) then
    cbc := Taur*Isat/Vt*ExMin*(vbc/Vt - EMin + 1) + Capcjc
  else if (vbc/Vt > EMax) then
    cbc := Taur*Isat/Vt*ExMax*(vbc/Vt - EMax + 1) + Capcjc
  else
    cbc := Taur*Isat/Vt*exp(vbc/Vt) + Capcjc;
  if (vbe/Vt < EMin) then
    cbe := Tauf*Isat/Vt*ExMin*(vbe/Vt - EMin + 1) + Capcje
  else if (vbe/Vt > EMax) then
    cbe := Tauf*Isat/Vt*ExMax*(vbe/Vt - EMax + 1) + Capcje
  else
    cbe := Tauf*Isat/Vt*exp(vbe/Vt) + Capcje;

  Ci := -((ibe - ibc)*qbk - ibc/Br - cbc*dervbc - Ccs*derCv);
  Bi := -(ibe/Bf + ibc/Br + cbc*dervbc + cbe*dervbe);
  Ei := -Bi - Ci + Ccs*derCv;

end;

function TBJTPNP.getb: integer;
begin
   result:=b;
end;

function TBJTPNP.getBi: TSample;
begin
   result:=Bi;
end;

function TBJTPNP.getc: Integer;
begin
   result:= c;
end;

function TBJTPNP.getCi: TSample;
begin
   result:=Ci;
end;

function TBJTPNP.gete: Integer;
begin
   result:= e;
end;

function TBJTPNP.getEi: TSample;
begin
   result:=Ei;
end;

{ TPMOS }

constructor TPMOS.Create(D_terminal, G_terminal, S_terminal, B_terminal: integer);
begin
   D:=D_terminal-1;
   G:=G_terminal-1;
   S:=S_terminal-1;
   B:=B_terminal-1;
   W:=20.0e-6; // "Width";
   L:=6.0e-6; // "Length";
   Beta:=0.0105e-3; // "Transconductance parameter";
   Vt:=-1.0; // "Zero bias threshold voltage";
   K2:=0.41; // "Bulk threshold parameter";
   K5:=0.839; // "Reduction of pinch-off region";
   dW:=-2.5e-6; // "Narrowing of channel";
   dL:=-2.1e-6; // "Shortening of channel";
   RDS:=1.e+7; // "Drain-Source-Resistance";
end;

destructor TPMOS.Destroy;
begin

  inherited;
end;

procedure TPMOS.Eval(t: TSample; x, dxdt: TVec);
var Sv, Dv, Bv, Gv: TSample;
begin
  if D=datum then begin Dv:=0;  end
  else begin Dv:=x[D];  end;

  if S=datum then begin Sv:=0;  end
  else begin Sv:=x[S];  end;

  if B=datum then begin Bv:=0; end
  else begin Bv:=x[B]; end;

  if G=datum then begin Gv:=0;  end
  else begin Gv:=x[G]; end;


  if ((RDS < 1.e-20) and (RDS > -1.0e-20)) then
     gds := 1.e20
  else
     gds := 1/RDS;
  v := Beta*(W + dW)/(L + dL);
  if (Dv > Sv) then
    ud := Sv
    else
    ud := Dv;
  if (Dv > Sv) then
    us := Dv
  else
    us := Sv;
  uds := ud - us;
  if (Bv < us) then
    ubs := 0
    else
    ubs := Bv - us;
  ugst := (Gv - us - Vt + K2*ubs)*K5;
  if (ugst >= 0) then
    id := v*uds*gds
    else if (ugst < uds) then
    id := -v*uds*(ugst - uds/2 - gds)
    else
    id := -v*(ugst*ugst/2 - uds*gds);
  Gi := 0;
  if (Dv > Sv) then
    Di := -id
    else
    Di := id;
  if (Dv > Sv) then
    Si := id
    else
    Si := -id;
  Bi := 0;
end;

function TPMOS.getBi: TSample;
begin
   result:=Bi;
end;

function TPMOS.getDi: TSample;
begin
   result:=Di;
end;

function TPMOS.getGi: TSample;
begin
   result:=Gi;
end;

function TPMOS.getSi: TSample;
begin
   result:=Si;
end;

{ TNMOS }

constructor TNMOS.Create(D_terminal, G_terminal, S_terminal,
  B_terminal: integer);
begin
   D:=D_terminal-1;
   G:=G_terminal-1;
   S:=S_terminal-1;
   B:=B_terminal-1;
   W:=20.0e-6; // "Width";
   L:=6.0e-6; // "Length";
   Beta:=0.0105e-3; // "Transconductance parameter";
   Vt:=-1.0; // "Zero bias threshold voltage";
   K2:=0.41; // "Bulk threshold parameter";
   K5:=0.839; // "Reduction of pinch-off region";
   dW:=-2.5e-6; // "Narrowing of channel";
   dL:=-2.1e-6; // "Shortening of channel";
   RDS:=1.e+7; // "Drain-Source-Resistance";
end;

destructor TNMOS.Destroy;
begin

  inherited;
end;

procedure TNMOS.Eval(t: TSample; x, dxdt: TVec);
var Sv, Dv, Bv, Gv: TSample;
begin
  if D=datum then begin Dv:=0;  end
  else begin Dv:=x[D];  end;

  if S=datum then begin Sv:=0;  end
  else begin Sv:=x[S];  end;

  if B=datum then begin Bv:=0;  end
  else begin Bv:=x[B];  end;

  if G=datum then begin Gv:=0;  end
  else begin Gv:=x[G];  end;

  if ((RDS < 1.0e-20) and (RDS > -1.0e-20)) then gds:= 1.e20 else gds:= 1/RDS;
  v := Beta*(W + dW)/(L + dL);
  if (Dv < Sv) then ud := Sv else ud := Dv;
  if (Dv < Sv) then us := Dv else us := Sv;
  uds := ud - us;
  if (Bv > us) then ubs := 0 else ubs := Bv - us;
  ugst := (Gv - us - Vt + K2*ubs)*K5;
  if (ugst <= 0) then id := v*uds*gds else if (ugst > uds) then
    id := v*uds*(ugst - uds/2 + gds)
    else id := v*(ugst*ugst/2 + uds*gds);
  Gi := 0;
  if (Dv < Sv) then Di := -id else Di := id;
  if (Dv < Sv) then Si := id else Si := -id;
  Bi := 0;
end;

function TNMOS.getBi: TSample;
begin
   result:=Bi;
end;

function TNMOS.getDi: TSample;
begin
   result:=Di;
end;

function TNMOS.getGi: TSample;
begin
   result:=Gi;
end;

function TNMOS.getSi: TSample;
begin
   result:=Si;
end;

{ TOPAMP }

constructor TOPAMPIDEAL.Create(in_pos, in_neg,
  out_terminal: Integer);
begin
  //Slope:=1; // "Slope of the out.v/vin characteristic at vin=0";
  p:= in_pos-1; // "Positive pin of the input port";
  n:= in_neg-1; // "Negative pin of the input port";
  o:= out_terminal-1; // "Output pin";

  {VMaxi := 0;
  VMini := 0;
  VMaxv:=15;
  VMinv:=-15;  }
end;

procedure TOPAMPIDEAL.Eval(t: TSample; x, dxdt: TVec);
begin
  {if p=datum then begin in_pv:=0;  end
  else begin in_pv:=x[p]; end;

  if n=datum then begin in_nv:=0;  end
  else begin in_nv:=x[n]; end;

  vin := in_pv - in_nv;
  f := 2/(VMaxv - VMinv);
  if (Slope < 0) then absSlope := -Slope else absSlope := Slope;
  if (f*vin < 0) then
   outv := (VMaxv + VMinv)/2 + absSlope*vin/(1 + absSlope*(-f*vin))
  else
   outv := (VMaxv + VMinv)/2 + absSlope*vin/(1 + absSlope*(f*vin));}

end;

function TOPAMPIDEAL.getinn: Integer;
begin
  result:= n;
end;

function TOPAMPIDEAL.getinp: Integer;
begin
  result:= p;
end;

function TOPAMPIDEAL.getout: Integer;
begin
  result:= o;
end;

{ TOPAMP }

constructor TOPAMP.Create(in_pos, in_neg,
  out_terminal: Integer);
begin
  Slope:=1000; // "Slope of the out.v/vin characteristic at vin=0";
  p:= in_pos-1; // "Positive pin of the input port";
  n:= in_neg-1; // "Negative pin of the input port";
  o:= out_terminal-1; // "Output pin";

  IMaxo := 0;
  IMino := 0;
  VMaxo:=14.5;
  VMino:=-14.2;
end;

procedure TOPAMP.Eval(t: TSample; x, dxdt: TVec);
begin
  vin:= 0.0;
  if p<>datum then
     vin:=x[p];
  if n<>datum then
     vin:=vin-x[n];

  f := 2/(VMaxo - VMino);
  if (Slope < 0) then
    absSlope := -Slope
  else
    absSlope := Slope;

  if (f*vin < 0) then
   outv := (VMaxo + VMino)/2 + absSlope*vin/(1 + absSlope*(-f*vin))
  else
   outv := (VMaxo + VMino)/2 + absSlope*vin/(1 + absSlope*(f*vin));
end;

function TOPAMP.getinn: Integer;
begin
  result:= n;
end;

function TOPAMP.getinp: Integer;
begin
  result:= p;
end;

function TOPAMP.getout: Integer;
begin
  result:= o;
end;

function TOPAMP.getV: TSample;
begin
   result:= outv;
end;

{ TVSinSource }

constructor TVSinSource.Create(p_terminal, n_terminal: integer; Amp,
  Freq: TSample);
begin
  A:=Amp;
  frec:=Freq;
  p:=p_terminal-1;
  n:=n_terminal-1;
end;

destructor TVSinSource.destroy;
begin
  inherited;
end;

procedure TVSinSource.Eval(t: TSample; x, dxdt: TVec);
begin
  v:=A*sin(2*pi*frec*t)
end;

function TVSinSource.getV: TSample;
begin
  result:=v;
end;

function TVSinSource.getn: Integer;
begin
  result:= n;
end;

function TVSinSource.getp: Integer;
begin
  result:= p;
end;

{ TChuaResistor }

constructor TChuaResistor.Create(pos_terminal, neg_terminal: integer;
  ValueGa, ValueGB, ValueVe: TSample);
begin
   Ga:=ValueGa;
   Gb:=ValueGb;
   Ve:=ValueVe;
   p:=pos_terminal-1;
   n:=neg_terminal-1;
end;

destructor TChuaResistor.destroy;
begin

  inherited;
end;

procedure TChuaResistor.Eval(t: TSample; x, dxdt: TVec);
begin
  v:=0.0;
  if (p<>Datum) then
      v:=x[p];
  if n<>Datum then
      v:=v-x[n];
  if (v < -Ve) then i := Gb*(v + Ve) - Ga*Ve
  else if (v > Ve) then
  i := Gb*(v - Ve) + Ga*Ve
  else i := Ga*v;
end;

function TChuaResistor.getI: TSample;
begin
  result:=i;
end;

function TChuaResistor.getn: Integer;
begin
  result:= n;
end;

function TChuaResistor.getp: Integer;
begin
  result:= p;
end;

{ TBJTPNPF }

constructor TBJTPNPF.Create(b_terminal, c_terminal, e_terminal: integer);
begin
  b:=b_terminal-1;
  c:=c_terminal-1;
  e:=e_terminal-1;
  isat := 1e-16;
	bf := 100.0;
  nf := one;
	vaf := zero;
	ikf := zero;
	ise := zero;
	ne := 1.5;
  br := one;
	nr := one;
	varr := zero;
	ikr := zero;
	isc := zero;
	nc := 2.0;
	re := zero;
	rb := zero;
	rbm := zero;
	irb := zero;
	rc := zero;
	eg := 1.11;
	cje := zero;
	vje := 0.75;
	mje := 0.33;
	cjc := zero;
	vjc := 0.75;
	mjc := 0.33;
	xcjc := one;
	fc := 0.5;
	tf := zero;
	xtf := zero;
	vtf := zero;
	itf := zero;
	tr := zero;
	xtb := zero;
	xti := 3.0;
	tre1 := zero;
	tre2 := zero;
	trb1 := zero;
	trb2 := zero;
	trm1 := zero;
	trm2 := zero;
	trc1 := zero;
	trc2 := zero;
	tnom := 300.0;
	t := 300.0;
	cjs := zero;
	mjs := zero;
	vjs := 0.75;
	area := one;
	ns := one;
	iss := zero;

	setNumTerms(4);
end;

destructor TBJTPNPF.Destroy;
begin
  inherited;
end;

procedure TBJTPNPF.Eval(t: TSample; x, dxdt: TVec);
begin
  if b=datum then begin Bv:=0.0; dBv:=0.0; end
  else begin Bv:=x[b]; dBv:=dxdt[b]; end;
  if c=datum then begin Cv:=0.0; dCv:=0.0; end
  else begin Cv:=x[c]; dCv:=dxdt[c]; end;
  if e=datum then begin Ev:=0.0; dEv:=0.0; end
  else begin Ev:=x[e]; dEv:=dxdt[e]; end;
  vbc := Bv - Cv;
  vbe := Bv - Ev;
  dvbcdt:= dBv-dCv;
  dvbedt:= dBv-dEv;
  dvCdt:= dCv;
  Vcjs:=Cv;
  dvcjsdt:=dvCdt;
  d2vbedt:=0.0;
  d2vbcdt:=0.0;
  //x[vbe,vbc,vcjs,dvbe,dvbc,dvcjs,d2vbe,d2vbc,d2vcjs]
	//x[0] : Vbe   Vbe
	//x[1] : Vbc   Vbc
	//x[2] : Vcjs
	//x[3] : dvbe/dt  dvbedt
	//x[4] : dvbc/dt  dvbcdt
	//x[5] : dvcjs/dt
	//x[6] : d2vbe/dt
	//x[7] : d2vbc/dt
	//x[8] : d2vcjs/dt
	//vp[0] : Vcjs ip[0] : Ic
	//vp[1] : Vbjs ip[1] : Ib
	//vp[2] : Vejs ip[2] : Ie
  vth := (kBoltzman * tnom) / eCharge;
  if (rbm<>0) then
    rbm := rb;

  //dc current equations

	Ibf := isat * (exp(Vbe/nf/vth) - one);
	Ile := ise * (exp(Vbe/ne/vth) - one);
	Ibr := isat * (exp(Vbc/nr/vth) - one);
	Ilc := isc * (exp(Vbc/nc/vth) - one);
	IIbe := Ibf / bf + Ile;
	IIbc := Ibr / br + Ilc;

  Ibf1 := (IIbe - Ile) * bf;
	Ibr1 := (IIbc - Ilc) * br;
	kqbtem := zero;

  if (ikf<>0) then
		kqbtem :=  4.0*(Ibf1/ikf);
	if (ikr<>0) then
		kqbtem := kqbtem + 4.0*(Ibr1/ikr);
	kqbtemp := sqrt(one + kqbtem);

  tempvaf := zero;
  if (vaf<>0) then
		tempvaf := Vbc / vaf;
  tempvar := zero;
	if (varr<>0) then
		tempvar := Vbe / varr;

  kqb := 0.5 * (one / (one - tempvaf - tempvar)) * (one + kqbtemp);
	IIce := (Ibf1 - Ibr1) / kqb;

  // Charge base-collector
	condassign(cbcj,Vbc-fc*vjc,
	cjc*power(one-fc,-one-mjc)*(one-fc*(one+mjc)+mjc*Vbe/vjc),
	cjc*power(one-Vbc/vjc,-mjc));
	cbct:=tr*isat*exp(Vbc/nr/vth)/(nr*vth);
	ibc:=area*(cbct+xcjc*cbcj)*dvbcdt;
	//end

  //Current Base-Emitter
	condassign(cbej,Vbe-fc*vje,
	cjc*power(one-fc,-one-mjc)*(one-fc*(one+mje)+mje*Vbe/vje),
	cje*power(one-Vbe/vje,-mje));

  tZ := one;
	if (vtf<>0) then
		tZ := exp(0.69*Vbc/vtf);

	cbet:=isat*exp(Vbe/nf/vth)*tf*(one+xtf*tZ)/(nf*vth);
  ibe:=area*(cbet+cbej)*dvbedt;

  //end

  Tibe := IIbe + ibe;
	Tibc := IIbc + ibc;

  rb1 := (rbm + (rb - rbm) / kqb) / area;
	vbx := Vbc + (Tibe + Tibc) * rb;

  condassign(cbx,vbx-fc*vjc,
	(one-xcjc)*cjc*power(one-fc,-one-mjc)*(one-fc*(one+mjc)+mjc*vbx/vjc),
	(one-xcjc)*cjc*power(one-vbx/vjc,-mjc));

  dibc := area*(cbct+xcjc*cbcj) * d2vbcdt;
	dibe := area*(cbet+cbej) * d2vbedt;
	dIIbe := isat/bf * exp(Vbe/nf/vth) * (dvbedt/nf/vth)
	+ ise * exp(Vbe/ne/vth) * (dvbedt/ne/vth);
	dIIbc := isat/br * exp(Vbe/nr/vth) * (dvbcdt/nr/vth)
	+ isc * exp(Vbc/nc/vth) * (dvbcdt/nc/vth);

	ibx := cbx * (dvbedt + rb*(dibc+dibe+dIIbc+dIIbe));

  IIjs := area * iss * (exp(Vcjs/ns/vth) - one);

  condassign(cjjs, Vcjs - zero,
	cjs*(one+mjs*Vcjs/vjs),
	cjs*power(one-Vcjs/vjs,-mjs));
	ijs := IIjs + area*(cjjs * dvcjsdt);

  Ci := - IIce + Tibc + ibx + ijs; // collector current
	Bi := - ibx - Tibc - Tibe;      // base current
	Ei := (IIce + Tibe - ijs);    // emitter current

	{vp[0] := x[2] + rc*ip[0];                // Collector Substrate Voltage
	vp[1] := x[2] + x[1] + ip[1]*rb;        // Base Substrate Voltage
	vp[2] := x[1] - x[0] + x[2] + ip[2]*re; // Emitter Substrate Voltage}
end;

function TBJTPNPF.getb: integer;
begin
  result:= b;
end;

function TBJTPNPF.getBi: TSample;
begin
   Result:=Bi;
end;

function TBJTPNPF.getc: Integer;
begin
   result:= c;
end;

function TBJTPNPF.getCi: TSample;
begin
   Result:=Ci;
end;

function TBJTPNPF.gete: Integer;
begin
   result:= e;
end;

function TBJTPNPF.getEi: TSample;
begin
   Result:=Ei;
end;


{ TDiodeZener }

constructor TDiodeZener.Create(p_terminal, n_terminal: Integer;
  Vz: TSample);
begin
  p:= p_terminal-1;
  n:= n_terminal-1;
  Vzener:= Vz;
  Gd:=5;
end;

destructor TDiodeZener.Destroy;
begin
  inherited;
end;

procedure TDiodeZener.Eval(t: TSample; x, dxdt: TVec);
var v: TSample;
begin
   v:=0.0;
   if (p<>Datum) then
      v:=x[p];
   if (n<>Datum) then
      v:=v-x[n];
   Iz:=0.5*Gd*(abs(v-vzener)+v-vzener);
end;

function TDiodeZener.getI: TSample;
begin
  result:= Iz;
end;

function TDiodeZener.getn: Integer;
begin
  result:= n;
end;

function TDiodeZener.getp: Integer;
begin
  result:= p;
end;

{ TDCVSource }

constructor TDCVSource.Create(p_terminal, n_terminal: integer;
  Voltage: TSample);
begin
  v:= Voltage;
  p:= p_terminal-1;
  n:= n_terminal-1;
end;

destructor TDCVSource.Destroy;
begin
  inherited;
end;

procedure TDCVSource.Eval(t: TSample; x, dxdt: TVec);
begin
  //v:=Voltage
end;

function TDCVSource.getV: TSample;
begin
   result:=v;
end;

function TDCVSource.getn: Integer;
begin
  result:= n;
end;

function TDCVSource.getp: Integer;
begin
  result:= p;
end;

{ TVABSSinSource }

constructor TVABSSinSource.Create(p_terminal, n_terminal: integer; Amp,
  Freq: TSample);
begin
  A:=Amp;
  frec:=Freq;
  p:=p_terminal-1;
  n:=n_terminal-1;
end;

destructor TVABSSinSource.Destroy;
begin
  inherited;
end;

procedure TVABSSinSource.Eval(t: TSample; x, dxdt: TVec);
begin
  v:=A*abs(sin(2*pi*frec*t));
end;

function TVABSSinSource.getV: TSample;
begin
   result:= v;
end;

function TVABSSinSource.getn: Integer;
begin
  result:= n;
end;

function TVABSSinSource.getp: Integer;
begin
  result:= p;
end;

{ TSIMPLENPN }

constructor TBJTEMNPN.Create(b_terminal, c_terminal, e_terminal: integer);
begin
  b:=b_terminal-1;
  c:=c_terminal-1;
  e:=e_terminal-1;
  Io:= 14.36e-15;
  AF:= 0.99;
  AR:= 0.5;
  Gmin:= 1e-12;
end;

destructor TBJTEMNPN.Destroy;
begin
  inherited;
end;

function TBJTEMNPN.Iexp(v: TSample): TSample;
  begin
	 Result:=Io*(exp(38.78*v)-1);
  end;

function TBJTEMNPN.Ibc(vb,vc,ve: TSample): TSample;
 begin
  Result:=AF*(Iexp(vb-ve))-Iexp(vb-vc)+Gmin*(vc-vb);
 end;

function TBJTEMNPN.Ibe(vb,vc,ve: TSample): TSample;
 begin
  Result:=Iexp(vb-ve)-AR*(Iexp(vb-vc))+Gmin*(vb-ve);
 end;

procedure TBJTEMNPN.Eval(t: TSample; x, dxdt: TVec);
begin
  if b<>datum then
    Bv:=x[b]
  else Bv:=0.0;

  if c<>datum then
    Cv:=x[c]
  else Cv:=0.0;

  if e<>datum then
    Ev:=x[e]
  else Ev:=0.0;

  Ci:=Ibc(Bv,Cv,Ev);
  Ei:=Ibe(Bv,Cv,Ev);
  Bi:=Ei-Ci;

end;

function TBJTEMNPN.getBi: TSample;
begin
  result:= Bi;
end;

function TBJTEMNPN.getCi: TSample;
begin
  RESULT:= Ci;
end;

function TBJTEMNPN.getEi: TSample;
begin
  result:= Ei;
end;

function TBJTEMNPN.getb: integer;
begin
  result:= b;
end;

function TBJTEMNPN.getc: Integer;
begin
  result:= c;
end;

function TBJTEMNPN.gete: Integer;
begin
  result:= e;
end;

{ TSIMPLEPNP }

constructor TBJTEMPNP.Create(b_terminal, c_terminal, e_terminal: integer);
begin
  b:=b_terminal-1;
  c:=c_terminal-1;
  e:=e_terminal-1;
  Io:= 1e-9;
  AF:= 0.9;
  AR:= 0.4;
  Gmin:= 0.0;
end;

destructor TBJTEMPNP.Destroy;
begin
  inherited;
end;

function TBJTEMPNP.Iexp(v: TSample): TSample;
  begin
	 Result:=Io*(exp(38.78*v)-1);
  end;

function TBJTEMPNP.Ibc(vb,vc,ve: TSample): TSample;
 begin
  Result:=AF*(Iexp(ve-vb))-Iexp(vc-vb)+Gmin*(vb-vc);
 end;

function TBJTEMPNP.Ibe(vb,vc,ve: TSample): TSample;
 begin
  Result:=Iexp(ve-vb)-AR*(Iexp(vc-vb));
 end;

procedure TBJTEMPNP.Eval(t: TSample; x, dxdt: TVec);
begin
  if b<>datum then
    Bv:=x[b]
  else Bv:=0.0;

  if c<>datum then
    Cv:=x[c]
  else Cv:=0.0;

  if e<>datum then
    Ev:=x[e]
  else Ev:=0.0;

  Ci:=Ibc(Bv,Cv,Ev);
  Ei:=Ibe(Bv,Cv,Ev);
  Bi:=Ei-Ci;

end;

function TBJTEMPNP.getBi: TSample;
begin
  result:= Bi;
end;

function TBJTEMPNP.getCi: TSample;
begin
  RESULT:= Ci;
end;

function TBJTEMPNP.getEi: TSample;
begin
  result:= Ei;
end;


function TBJTEMPNP.getb: integer;
begin
  result:= b;
end;

function TBJTEMPNP.getc: Integer;
begin
  result:= c;
end;

function TBJTEMPNP.gete: Integer;
begin
  result:= e;
end;

{ TDCISource }

constructor TDCISource.Create(p_terminal, n_terminal: integer;
  Current: TSample);
begin
  p:= p_terminal-1;
  n:= n_terminal-1;
  I:= Current;
end;

destructor TDCISource.Destroy;
begin
  inherited;
end;

procedure TDCISource.Eval(t: TSample; x, dxdt: TVec);
begin
  //
end;

function TDCISource.getI: TSample;
begin
   result:= I;
end;

function TDCISource.getn: Integer;
begin
  result:= n;
end;

function TDCISource.getp: Integer;
begin
  result:= p;
end;

{ TVCT }

constructor TVCT.Create(inpos_terminal, inneg_terminal, outpos_terminal,
  outneg_terminal: Integer; G: TSample);
begin
  pin:= inpos_terminal-1;
  nin:= inneg_terminal-1;
  pout:= outpos_terminal-1;
  nout:= outneg_terminal-1;
  Gtran:= G;
end;

procedure TVCT.Eval(t: TSample; x, dxdt: TVec);
var v: TSample;
begin
   v:= 0.0;
   if pin<>Datum then
    v:= x[pin];
   if nin<>Datum then
    v:=v-x[nin];
   I:= Gtran*v;
end;

function TVCT.getI: TSample;
begin
  result:= I;
end;

function TVCT.getinn: Integer;
begin
  result:= nin;
end;

function TVCT.getinp: Integer;
begin
  result:= pin;
end;

function TVCT.getoutn: Integer;
begin
  result:= nout;
end;

function TVCT.getoutp: Integer;
begin
  result:= pout;
end;

{ TVVT }

constructor TVVT.Create(inpos_terminal, inneg_terminal, outpos_terminal,
  outneg_terminal: Integer; Mu: TSample);
begin
  pin:= inpos_terminal-1;
  nin:= inneg_terminal-1;
  pout:= outpos_terminal-1;
  nout:= outneg_terminal-1;
  Mutran:= Mu;
end;

procedure TVVT.Eval(t: TSample; x, dxdt: TVec);
var vi: TSample;
begin
   vi:= 0.0;
   if pin<>Datum then
    vi:= x[pin];
   if nin<>Datum then
    vi:=vi-x[nin];
   V:= Mutran*vi;
end;

function TVVT.getinn: Integer;
begin
   result:= nin;
end;

function TVVT.getinp: Integer;
begin
   result:= pin;
end;

function TVVT.getoutn: Integer;
begin
   result:= nout;
end;

function TVVT.getoutp: Integer;
begin
   result:= pout;
end;

function TVVT.getV: TSample;
begin
   result:= V;
end;

{ TOPAMP741 }

constructor TOPAMP741.Create(inp, inn, output, vcc, vee: integer);
begin
  p:= inp-1;
  n:= inn-1;
  o:= output-1;
  vc:= vcc-1;
  ve:= vee-1;
end;

procedure TOPAMP741.Eval(t: TSample; x, dxdt: TVec);
begin
  //
end;

function TOPAMP741.getinn: Integer;
begin
  result:= n;
end;

function TOPAMP741.getinp: Integer;
begin
  result:= p;
end;

function TOPAMP741.getout: Integer;
begin
  result:= o;
end;

function TOPAMP741.getvcc: Integer;
begin
  result:= vc;
end;

function TOPAMP741.getvee: Integer;
begin
  result:= ve;
end;

{ TBJTPWLNPN }

constructor TBJTPWLNPN.Create(b_terminal, c_terminal, e_terminal: integer);
begin
  b:=b_terminal-1;
  c:=c_terminal-1;
  e:=e_terminal-1;
  Gi:= 1/(1.0e3);
  Go:= 1000;
  Vbe:= 0.7;
  Vcesat:= 0.0;
  u:= 0.01;
  beta:= 100.0;
end;

destructor TBJTPWLNPN.Destroy;
begin
  inherited;
end;

procedure TBJTPWLNPN.Eval(t: TSample; x, dxdt: TVec);
begin
  if b<>datum then
    Bv:=x[b]
  else Bv:=0.0;

  if c<>datum then
    Cv:=x[c]
  else Cv:=0.0;

  if e<>datum then
    Ev:=x[e]
  else Ev:=0.0;

  Bi:=Ibe(Bv,Cv,Ev);
  Ci:=Ice(Bv,Cv,Ev);
  Ei:=Bi+Ci;
end;

function TBJTPWLNPN.getb: integer;
begin
  result:= b;
end;

function TBJTPWLNPN.getBi: TSample;
begin
  result:= Bi;
end;

function TBJTPWLNPN.getc: Integer;
begin
  result:= c;
end;

function TBJTPWLNPN.getCi: TSample;
begin
  RESULT:= Ci;
end;

function TBJTPWLNPN.gete: Integer;
begin
  result:= e;
end;

function TBJTPWLNPN.getEi: TSample;
begin
  result:= Ei;
end;

function TBJTPWLNPN.Ibe(vb, vc, ve: TSample): TSample;
begin
  result:= 0.5*Gi*(abs(vb-ve)-(Vbe-u*(Vc-Ve))+(vb-ve)-(Vbe-u*(Vc-Ve)));
end;

function TBJTPWLNPN.Ice(vb, vc, ve: TSample): TSample;
begin
  result:= -0.5*Go*(abs(ve-vc)-Vcesat+(ve-vc)-Vcesat);
  result:= result+beta*Bi;
end;

{ TBJTPWLPNP }

constructor TBJTPWLPNP.Create(b_terminal, c_terminal, e_terminal: integer);
begin
  b:=b_terminal-1;
  c:=c_terminal-1;
  e:=e_terminal-1;
  Gi:= 1/(1.0e3);
  Go:= 1000;
  Vbe:= 0.7;
  Vcesat:= 0.0;
  u:= 0.01;
  beta:= 50.0;
end;

destructor TBJTPWLPNP.Destroy;
begin
  inherited;
end;

procedure TBJTPWLPNP.Eval(t: TSample; x, dxdt: TVec);
begin
  if b<>datum then
    Bv:=x[b]
  else Bv:=0.0;

  if c<>datum then
    Cv:=x[c]
  else Cv:=0.0;

  if e<>datum then
    Ev:=x[e]
  else Ev:=0.0;

  Bi:=Ibe(Bv,Cv,Ev);
  Ci:=Ice(Bv,Cv,Ev);
  Ei:=Bi+Ci;

end;

function TBJTPWLPNP.getb: integer;
begin
  result:= b;
end;

function TBJTPWLPNP.getBi: TSample;
begin
  result:= Bi;
end;

function TBJTPWLPNP.getc: Integer;
begin
  result:= c;
end;

function TBJTPWLPNP.getCi: TSample;
begin
  result:= Ci;
end;

function TBJTPWLPNP.gete: Integer;
begin
  result:= e;
end;

function TBJTPWLPNP.getEi: TSample;
begin
  result:= Ei;
end;

function TBJTPWLPNP.Ibe(vb, vc, ve: TSample): TSample;
begin
  result:= 0.5*Gi*(abs(ve-vb)-(Vbe-u*(Ve-Vc))+(ve-vb)-(Vbe-u*(Ve-Vc)));
end;

function TBJTPWLPNP.Ice(vb, vc, ve: TSample): TSample;
begin
  result:= -0.5*Go*(abs(vc-ve)-Vcesat+(vc-ve)-Vcesat);
  result:= result+beta*Bi;
end;

{ TOPAMPMM }

constructor TOPAMPMM.Create(inp, inn, output, vcc, vee: integer);
begin
  p:= inp-1;
  n:= inn-1;
  o:= output-1;
  vc:= vcc-1;
  ve:= vee-1;
end;

procedure TOPAMPMM.Eval(t: TSample; x, dxdt: TVec);
begin
 //
end;

function TOPAMPMM.getinn: Integer;
begin
  result:= n;
end;

function TOPAMPMM.getinp: Integer;
begin
  result:= p;
end;

function TOPAMPMM.getout: Integer;
begin
  result:= o;
end;

function TOPAMPMM.getvcc: Integer;
begin
  result:= vc;
end;

function TOPAMPMM.getvee: Integer;
begin
  result:= ve;
end;

{ TOPAMPMacroModel }

constructor TOPAMPMacroModel.Create(inp, inn, output, vcc, vee: integer);
begin
  p:= inp-1;
  n:= inn-1;
  o:= output-1;
  vc:= vcc-1;
  ve:= vee-1;
end;

procedure TOPAMPMacroModel.Eval(t: TSample; x, dxdt: TVec);
begin
  //
end;

function TOPAMPMacroModel.getinn: Integer;
begin
  result:= n;
end;

function TOPAMPMacroModel.getinp: Integer;
begin
  result:= p;
end;

function TOPAMPMacroModel.getout: Integer;
begin
  result:= o
end;

function TOPAMPMacroModel.getvcc: Integer;
begin
  result:= vc
end;

function TOPAMPMacroModel.getvee: Integer;
begin
  result:= ve;
end;

end.
