unit DENewtonSolver;
// Differential Evolution Solver Class
// Based on algorithms developed by Dr. Rainer Storn & Kenneth Price
// Written By: Lester E. Godwin
//             PushCorp, Inc.
//             Dallas, Texas
//             972-840-0208 x102
//             godwin@pushcorp.com
// Created: 6/8/98
// Last Modified: 6/8/98
// Revision: 1.0
//
// Ported To Delphi From C++ May 22, 2007
// Ported To Delphi By: Gustavo Sanchez
//                      Nuevo Laredo, Tamaulipas, Mexico.
//                      jsanch85@prodigy.net.mx
//                      jsanch85@visteon.com
//  Differential Evolution with Newton Method
//                      May 28, 2007. 

interface

uses
  SysUtils, Math387, CuBLASMtxVec, Math;

const
  stBest1Exp = 0;
  stRand1Exp = 1;
  stRandToBest1Exp = 2;
  stBest2Exp = 3;
  stRand2Exp = 4;
  stBest1Bin = 5;
  stRand1Bin = 6;
  stRandToBest1Bin = 7;
  stBest2Bin = 8;
  stRand2Bin = 9;
  SEED = 3;
  IM1 = 2147483563;
  IM2 = 2147483399;
  IA1 = 40014;
  IA2 = 40692;
  IQ1 = 53668;
  IQ2 = 52774;
  IR1 = 12211;
  IR2 = 3791;
  NTAB = 32;
  EPS = 1.2e-7;
  AM= 1.0/IM1;
  IMM1= IM1-1;
  NDIV= 1+IMM1/NTAB;
  RNMX= 1.0-EPS;

var
  idum: longint;
  idum2: longint;
  iy: longint;
  iv: array[0..NTAB] of longint;



type
  TStrategyFunction = procedure (tipo : integer) of object;
  TDESolver = class(TObject)
  private
  idum,
  idum2,
  iy:  longint;
  iv: array[0..NTAB] of longint;
    procedure Best1Bin(candidate : integer);
    procedure Best1Exp(candidate : integer);
    procedure Best2Bin(candidate : integer);
    procedure Best2Exp(candidate : integer);
    procedure Rand1Bin(candidate : integer);
    procedure Rand1Exp(candidate : integer);
    procedure Rand2Bin(candidate : integer);
    procedure Rand2Exp(candidate : integer);
    procedure RandToBest1Bin(candidate : integer);
    procedure RandToBest1Exp(candidate : integer);
  protected
    bestEnergy: TSample;
    bestSolution: TVec;
    calcTrialSolution: TStrategyFunction;
    generations: Integer;
    nDim: Integer;
    nPop: Integer;
    popEnergy: TVec;
    population: TVec;
    probability: TSample;
    scale: TSample;
    strategy: Integer;
    trialEnergy: TSample;
    trialSolution: TVec;
    trialSolutionFunction: TVec;
    function RandomUniform(minValue, maxValue : TSample): TSample;
    procedure SelectSamples(candidate: integer; var r1 ,  r2 , r3 , r4 , r5 :
            Integer);
  public
    constructor Create(dim, popSize : integer);
    destructor Destroy; override;
    function Dimension: Integer;
    function Energy: TSample;
    function EnergyFunction(testSolution, VectorValue: TVec; var bAtSolution :
            boolean): TSample; virtual; abstract;
    function Generationes: Integer;
    function Populations: Integer;
    procedure Setup(min, max : TVec; deStrategy : integer; diffScale,
            crossoverProb : TSample);
    function Solution: TVec;
    function Solve(maxGenerations : integer): Boolean; virtual;
  end;

  TDENewtonSolver= Class(TDESolver)
  public
    constructor Create(dim, popSize: integer);
    destructor Destroy; override;
    procedure Newton(tolerance: TSample = 1e-7);
  private
    actualnorm: TSample;
    jacobian: TMtx;
    F0: TVec;
    xdelta, temp: TVec;
    tol: TSample
  end;



implementation

{
********************************** TDESolver ***********************************
}
constructor TDESolver.Create(dim, popSize : integer);
begin
  nDim:= dim;
  nPop:= popSize;
  generations:= 0;
  strategy:= stRand1Exp;
  scale:= 0.7;
  probability:= 0.5;
  bestEnergy:= 0.0;
  trialSolution:= TVec.Create;
  trialSolution.Size(nDim);
  bestSolution:= TVec.Create;
  bestSolution.Size(nDim);
  trialSolutionFunction:= TVec.Create;
  trialSolutionFunction.Size(nDim);
  popEnergy:= TVec.Create;
  popEnergy.Size(nPop);
  population:= TVec.Create;
  population.Size(nPop*nDim);
  idum2:= 123456789;
  iy:= 0;
end;

destructor TDESolver.Destroy;
begin
  trialSolution.Destroy;
  bestSolution.Destroy;
  popEnergy.Destroy;
  population.Destroy;
  inherited Destroy;
end;

procedure TDESolver.Best1Bin(candidate : integer);
var
  r1, r2, r3, r4, r5, i, n: Integer;
begin
  r1:= 1;
  r2:= 1;
  r3:= 0;
  r4:= 0;
  r5:= 0;
  SelectSamples(candidate, r1, r2, r3, r4, r5);
  n:= Round(RandomUniform(0.0, nDim-1));

  for i:=0 to nDim-1 do
    trialSolution[i]:= population[candidate*nDim+i];

  for i:=0 to nDim-1 do
   begin
    if (RandomUniform(0.0, 1.0) >= probability) then break;
    trialSolution[n]:=bestSolution[n]+
                      scale*(population[r1*nDim+n]
                      -population[r2*nDim+n]);
    n:= (n+1)div nDim;
   end;
end;

procedure TDESolver.Best1Exp(candidate : integer);
var
  r1, r2, r3, r4, r5, n, i: Integer;
begin
  r1:= 1;
  r2:= 1;
  r3:= 0;
  r4:= 0;
  r5:= 0;
  SelectSamples(candidate, r1, r2, r3, r4, r5);
  n:= Round(RandomUniform(0.0, nDim-1));

  for i:=0 to nDim-1 do
    trialSolution[i]:= population[candidate*nDim+i];

  for i:=0 to nDim-1 do
   begin
    if (RandomUniform(0.0, 1.0) >= probability) then break;
      trialSolution[n]:=bestSolution[n]+scale*(population[r1*nDim+n]
                               -population[r2*nDim+n]);
    n:= (n+1)div nDim;

   end;
end;

procedure TDESolver.Best2Bin(candidate : integer);
var
  r1, r2, r3, r4, r5, i, n: Integer;
begin
  r1:= 1;
  r2:= 1;
  r3:= 1;
  r4:= 1;
  r5:= 0;
  SelectSamples(candidate, r1, r2, r3, r4, r5);
  n:= Round(RandomUniform(0.0, nDim-1));

  for i:=0 to nDim-1 do
    trialSolution[i]:= population[candidate*nDim+i];

  for i:=0 to nDim-1 do
   begin
    if (RandomUniform(0.0, 1.0) >= probability) then break;
    trialSolution[n]:=bestSolution[n]
                     +scale*(population[r1*nDim+n]
                     +population[r2*nDim+n]
                     -population[r3*nDim+n]
                     -population[r4*nDim+n]);

    n:= (n+1)div nDim;
   end;
end;

procedure TDESolver.Best2Exp(candidate : integer);
var
  r1, r2, r3, r4, r5, i, n: Integer;
begin
  r1:= 1;
  r2:= 1;
  r3:= 1;
  r4:= 1;
  r5:= 0;
  SelectSamples(candidate, r1, r2, r3, r4, r5);
  n:= Round(RandomUniform(0.0, nDim-1));

  for i:=0 to nDim-1 do
    trialSolution[i]:= population[candidate*nDim+i];

  for i:=0 to nDim-1 do
   begin
    if (RandomUniform(0.0, 1.0) >= probability) then break;
    trialSolution[n]:=bestSolution[n]
                     +scale*(population[r1*nDim+n]
                     +population[r2*nDim+n]
                     -population[r3*nDim+n]
                     -population[r4*nDim+n]);
    n:= (n+1)div nDim;
   end;
end;

function TDESolver.Dimension: Integer;
begin
  result:= NDim;
end;

function TDESolver.Energy: TSample;
begin
  result:= bestEnergy;
end;

function TDESolver.Generationes: Integer;
begin
  result:= generations;
end;

function TDESolver.Populations: Integer;
begin
  result:= nPop;
end;

procedure TDESolver.Rand1Bin(candidate : integer);
var
  r1, r2, r3, r4, r5, i, n: Integer;
begin
  r1:= 1;
  r2:= 1;
  r3:= 1;
  r4:= 0;
  r5:= 0;
  SelectSamples(candidate, r1, r2, r3, r4, r5);
  n:= Round(RandomUniform(0.0, nDim-1));

  for i:=0 to nDim-1 do
    trialSolution[i]:= population[candidate*nDim+i];

  for i:=0 to nDim-1 do
   begin
    if (RandomUniform(0.0, 1.0) >= probability) then break;
    trialSolution[n]:=population[r1*nDim+n]
                     +scale*(population[r2*nDim+n]
                     -population[r3*nDim+n]);

    n:= (n+1)div nDim;
   end;
end;

procedure TDESolver.Rand1Exp(candidate : integer);
var
  r1, r2, r3, r4, r5, i, n: Integer;
begin
  r1:= 1;
  r2:= 1;
  r3:= 1;
  r4:= 0;
  r5:= 0;
  SelectSamples(candidate, r1, r2, r3, r4, r5);
  n:= Round(RandomUniform(0.0, nDim-1));

  for i:=0 to nDim-1 do
    trialSolution[i]:= population[candidate*nDim+i];

  for i:=0 to nDim-1 do
   begin
    if (RandomUniform(0.0, 1.0) >= probability) then break;
      trialSolution[n]:=population[r1*nDim+n]+
                      scale*(population[r2*nDim+n]
                      -population[r3*nDim+n]);
    n:= (n+1)div nDim;
   end;
end;

procedure TDESolver.Rand2Bin(candidate : integer);
var
  r1, r2, r3, r4, r5, i, n: Integer;
begin
  r1:= 1;
  r2:= 1;
  r3:= 1;
  r4:= 1;
  r5:= 1;
  SelectSamples(candidate, r1, r2, r3, r4, r5);
  n:= Round(RandomUniform(0.0, nDim-1));

  for i:=0 to nDim-1 do
    trialSolution[i]:= population[candidate*nDim+i];

  for i:=0 to nDim-1 do
   begin
    if (RandomUniform(0.0, 1.0) >= probability) then break;
    trialSolution[n]:=population[r1*nDim+n]
                     +scale*(population[r2*nDim+n]
                     +population[r3*nDim+n]
                     -population[r4*nDim+n]
                     -population[r5*nDim+n]);

    n:= (n+1)div nDim;
   end;
end;

procedure TDESolver.Rand2Exp(candidate : integer);
var
  r1, r2, r3, r4, r5, i, n: Integer;
begin
  r1:= 1;
  r2:= 1;
  r3:= 1;
  r4:= 1;
  r5:= 1;
  SelectSamples(candidate, r1, r2, r3, r4, r5);
  n:= Round(RandomUniform(0.0, nDim-1));

  for i:=0 to nDim-1 do
    trialSolution[i]:= population[candidate*nDim+i];

  for i:=0 to nDim-1 do
   begin
    if (RandomUniform(0.0, 1.0) >= probability) then break;
    trialSolution[n]:=population[r1*nDim+n]+
                      scale*(population[r2*nDim+n]
                      +population[r3*nDim+n]
                      -population[r4*nDim+n]
                      -population[r5*nDim+n]);
    n:= (n+1)div nDim;
   end;
end;

function TDESolver.RandomUniform(minValue, maxValue : TSample): TSample;
var
  j, k: Integer;
begin
  if (iy=0) then
    idum:= SEED;

  if (idum<=0) then
    begin
     if (-idum<1) then
       idum:= 1
     else
       idum:= -idum;

     idum2:= idum;

     for j:= NTAB+7 downto 0 do
      begin
       k:= round(idum/ IQ1);
       idum:= IA1*(idum - k*IQ1)-k*IR1;
       if (idum<0) then idum := idum + IM1;
       if (j < NTAB) then iv[j]:= idum;
      end;
    iy:= iv[0];
    end;

    k:= round(idum/IQ1);
    idum:= IA1*(idum-k*IQ1)-k*IR1;

    if (idum < 0) then
      idum := idum+IM1;

    k := round(idum2/IQ2);
    idum2:= IA2*(idum2-k*IQ2)-k*IR2;

    if (idum2 < 0) then
     idum2:= idum2+IM2;

    j:= round(iy/NDIV);
    iy:= iv[j] -idum2;
    iv[j]:= idum;

    if (iy < 1) then
       iy:= iy+IMM1;

    result:= AM*iy;

    if (result > RNMX) then
      result:= RNMX
    else
    result:= minValue + result*(maxValue - minValue);
end;

procedure TDESolver.RandToBest1Bin(candidate : integer);
var
  r1, r2, r3, r4, r5, i, n: Integer;
begin
  r1:= 1;
  r2:= 1;
  r3:= 0;
  r4:= 0;
  r5:= 0;
  SelectSamples(candidate, r1, r2, r3, r4, r5);
  n:= Round(RandomUniform(0.0, nDim-1));

  for i:=0 to nDim-1 do
    trialSolution[i]:= population[candidate*nDim+i];

  for i:=0 to nDim-1 do
   begin
    if (RandomUniform(0.0, 1.0) >= probability) then break;
    trialSolution[n]:=trialSolution[n]+
                     scale*(bestSolution[n]-trialSolution[n])
                     +scale*(population[r1*nDim+n]
                     -population[r2*nDim+n]);
    n:= (n+1)div nDim;
   end;
end;

procedure TDESolver.RandToBest1Exp(candidate : integer);
var
  r1, r2, r3, r4, r5, i, n: Integer;
begin
  r1:= 1;
  r2:= 1;
  r3:= 0;
  r4:= 0;
  r5:= 0;
  SelectSamples(candidate, r1, r2, r3, r4, r5);
  n:= Round(RandomUniform(0.0, nDim-1));

  for i:=0 to nDim-1 do
    trialSolution[i]:= population[candidate*nDim+i];

  for i:=0 to nDim-1 do
   begin
    if (RandomUniform(0.0, 1.0) >= probability) then break;
      trialSolution[n]:=trialSolution[n]+
                     scale*(bestSolution[n]-trialSolution[n])
                     +scale*(population[r1*nDim+n]
                     -population[r2*nDim+n]);
    n:= (n+1)div nDim;
   end;
end;

procedure TDESolver.SelectSamples(candidate: integer; var r1 ,  r2 , r3 , r4 ,
        r5 : Integer);
begin
  if (r1<>0) then
   begin
    repeat
      r1:=Round(RandomUniform(0.0, nPop-1));
    until not(r1 = candidate);
   end;
  if (r2<>0) then
   begin
    repeat
      r2:=Round(RandomUniform(0.0, nPop-1));
    until (not(r2 = candidate) and not(r2 =r1));
   end;
  if (r3<>0) then
   begin
    repeat
      r3:=Round(RandomUniform(0.0, nPop-1));
    until (not(r3 = candidate) and not(r3 = r2) and not(r3 = r1));
   end;
  if (r4<>0) then
   begin
    repeat
      r4:=Round(RandomUniform(0.0, nPop-1));
    until (not(r4 = candidate) and not(r4 = r3) and not(r4 = r2) and not(r4 = r1));
   end;
  if (r5<>0) then
   begin
    repeat
      r5:=Round(RandomUniform(0.0, nPop-1));
    until (not(r5 = candidate) and not(r5 = r4) and not(r5 = r3) and not(r5 = r2)
           and not(r5 = r1));
   end;
end;

procedure TDESolver.Setup(min, max : TVec; deStrategy : integer;
        diffScale, crossoverProb : TSample);
var
  i, j: Integer;
begin
    strategy:= deStrategy;
    scale:= diffScale;
    probability:= crossoverProb;

    for i:=0 to nPop-1 do
     begin
      for j:=0 to nDim-1 do
      begin
        population[i*nDim+j] := RandomUniform(min[j],max[j]);
        popEnergy[i]:= 1.0E20;
      end;
     end;

    for i:=0 to nDim-1 do
     bestSolution[i]:= 0.0;

    case strategy of
      0:
          calcTrialSolution := Best1Exp;
      1:
          calcTrialSolution := Rand1Exp;
      2:
          calcTrialSolution := RandToBest1Exp;
      3:
          calcTrialSolution := Best2Exp;
      4:
          calcTrialSolution := Rand2Exp;
      5:
          calcTrialSolution := Best1Bin;
      6:
          calcTrialSolution := Rand1Bin;
      7:
          calcTrialSolution := RandToBest1Bin;
      8:
          calcTrialSolution := Best2Bin;
      9:
          calcTrialSolution := Rand2Bin;
  end;
end;

function TDESolver.Solution: TVec;
begin
  result:= bestSolution;
end;

function TDESolver.Solve(maxGenerations : integer): Boolean;
var
  generation, candidate: Integer;
  bAtSolution: Boolean;
  i: Integer;
begin
  bestEnergy:= 1.0E20;
  bAtSolution:= False;

  for generation:=0 to maxGenerations-1 do
   begin
     if (bAtSolution) then break;
     for candidate:=0 to nPop-1 do
     begin
      calcTrialSolution(candidate);
      trialEnergy:= EnergyFunction(trialSolution, trialSolutionFunction, bAtSolution);
      if (trialenergy < popEnergy[candidate]) then
      begin
        popEnergy[candidate]:= trialEnergy;
        for i:=0 to nDim-1 do
           population[candidate*nDim+i]:=trialSolution[i];
        if (trialEnergy < bestEnergy) then
        begin
          bestEnergy := trialEnergy;
          for i:=0 to nDim-1 do
            bestSolution[i]:=trialSolution[i];
        end;
      end;
     end;
   end;
   generations:= generation;
   result:= bAtSolution;
end;

{ TDENewtonSolver }

constructor TDENewtonSolver.Create(dim, popSize: integer);
begin
   inherited Create(dim, popSize);
   jacobian:= TMtx.Create;
   jacobian.Size(dim, dim);
   F0:= TVec.Create;
   F0.Size(dim);
   xdelta:= TVec.Create;
   xdelta.Size(dim);
   temp:= TVec.Create;
   temp.Size(nDim);
end;

destructor TDENewtonSolver.Destroy;
begin
   jacobian.Destroy;
   F0.Destroy;
   xdelta.Destroy;
   temp.Destroy;
   inherited Destroy;
end;

procedure TDENewtonSolver.Newton(tolerance: TSample = 1e-7);
const
   delta= 1e-7;
   invdelta= 1/delta;
   max_itt = 3000;
var
   i, itt: integer;
   bAtSolution: Boolean;
begin
   tol:= tolerance;
   itt:= 0;
   while (bestEnergy > tol) {and (itt < max_itt)} do
   begin
    inc(itt);
      for i:=0 to nDim-1 do
      begin
      trialSolution.Assign(bestSolution);
      trialSolution[i]:= trialSolution[i]+delta;
      EnergyFunction(trialSolution, F0, bAtSolution);
      F0.Sub(trialSolutionFunction);
      jacobian.SetCol(F0,i);
      end;
      jacobian.AdaptiveSGMRESK(trialSolutionFunction, xdelta, 0);
      xdelta.Scale(delta);
      for i:=0 to nDim-1 do
         if abs(xdelta[i]) > 0.25 then
             xdelta[i]:= 0.25*Math.sign(xdelta[i]);
      i:=1;
      trialSolution.Assign(bestSolution);
      trialSolution.Sub(xdelta);
      actualnorm:= EnergyFunction(trialSolution, F0, bAtSolution);
      while ((actualnorm > bestenergy) and (i<4)) do
      begin
      temp.Assign(xdelta);
      temp.Scale(exp(-i));
      trialSolution.Assign(bestSolution);
      trialSolution.Sub(temp);
      actualnorm:= EnergyFunction(trialSolution, F0, bAtSolution);
      inc(i);
      end;
      bestSolution.Assign(trialSolution);
      bestEnergy:= EnergyFunction(bestSolution, trialSolutionFunction, bAtSolution);
   end
end;

end.

